using ImageSharp.Images;
using ImageSharpApp.AlgProcessor.Forms;
using System.Windows.Forms;
using ImageSharp;
using ImageSharp.GrayValueTransforms;
namespace ImageSharpApp.AlgProcessor
{
    abstract class AlgGrayImgProcessor : IAlgProcessor
    {
        public bool Apply(IImage img)
        {
            if (!(img is ImageGray))
            {
                return false;
            }
            else
            {
                return this.Apply(img as ImageGray);
            }
        }

        public abstract bool Apply(ImageGray img);
    }

    class GeoSeamCarvingProc : IAlgProcessor
    {
        public bool Apply(IImage img)
        {
            if (!(img is ImageGray) && !(img is ImageColor))
            {
                return false;
            }

            GeoSeamCarvingParams pars = new GeoSeamCarvingParams(img.Width, img.Height);
            GeoSeamCarvingForm form = new GeoSeamCarvingForm(pars, img.Width, img.Height);
            if (form.ShowDialog() == DialogResult.OK)
            {
                if (img is ImageGray)
                {
                    ((ImageGray)img).Apply(new SeamCarving(pars.NWidth, pars.NHeight));
                }
                else if (img is ImageColor)
                {
                    ((ImageColor)img).Apply(new SeamCarving(pars.NWidth, pars.NHeight));
                }
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class GrayContrastAdaptProc : AlgGrayImgProcessor
    {
        GrayContrastAdaptParams pars = new GrayContrastAdaptParams();
        
        public override bool Apply(ImageGray img)
        {
            GrayContrastAdaptForm form = new GrayContrastAdaptForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ContrastAdaption(pars.PxPercLow, pars.PxPercHigh));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class GrayContrastStretchProc : AlgGrayImgProcessor
    {
        GrayContrastStretchParams pars = new GrayContrastStretchParams();
        
        public override bool Apply(ImageGray img)
        {
            GrayContrastStretchForm form = new GrayContrastStretchForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.GrayValueTransforms.ContrastStretching(pars.GrayOld1, pars.GrayNew1, pars.GrayOld2, pars.GrayNew2));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class GrayGammaCorrectionProc : AlgGrayImgProcessor
    {
        GrayGammaCorrectionParams pars = new GrayGammaCorrectionParams();

        public override bool Apply(ImageGray img)
        {
            GrayGammaCorrectionForm form = new GrayGammaCorrectionForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.GrayValueTransforms.GammaCorrection(pars.Gamma, pars.A0));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class GrayLUTApplicationProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            GrayLUTParams pars = new GrayLUTParams();
            for (int i = 0; i < pars.LUT.Length; i++)
            {
                pars.LUT[i] = (byte)(255 - i * i / pars.LUT.Length);
            }
            if (img != null)
            {
                img.Apply(new ImageSharp.GrayValueTransforms.LUTApplication(pars.LUT));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class GrayFunctionApplicationProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            GrayFunctionParams pars = new GrayFunctionParams();
            if (img != null)
            {
                img.Apply(new ImageSharp.GrayValueTransforms.FunctionApplication(pars.TransFunc));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class GrayImgInversionProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.GrayValueTransforms.GrayValueInversion());
            return true;
        }
    }

    class GrayHistEqualizationProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.HistOps.HistEqualization());
            return true;
        }
    }

    class GrayHistLocalEqualizationProc : AlgGrayImgProcessor
    {
        GrayHistLocalEqParams pars = new GrayHistLocalEqParams();

        public override bool Apply(ImageGray img)
        {
            GrayHistLocalEqForm form = new GrayHistLocalEqForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.HistOps.HistLocalEqualization(pars.MaskWidth));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class GrayHistSpecificationProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            GrayHistSpecParams pars = new GrayHistSpecParams();
            if (img != null)
            {
                //img.Apply(new ImageSharp.HistOps.HistSpecification(pars.Hist));
                img.Apply(new ImageSharp.HistOps.HistSpecGauss(50.0f));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class SegThresholdingProc : AlgGrayImgProcessor
    {
        SegThreshParams pars = new SegThreshParams();

        public override bool Apply(ImageGray img)
        {
            SegThresholdingForm form = new SegThresholdingForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Segmentations.Binarizations.Thresholding(pars.Threshold));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class SegOtsuThresholdingProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.Segmentations.Binarizations.OtsuThresholding());
            return true;
        }
    }

    class SegADBProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            SegADBParams pars = new SegADBParams();
            if (img != null)
            {
                img.Apply(new ImageSharp.Segmentations.Binarizations.ADBinarization(pars.GCells));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterLinearGaussProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(ImageSharp.Filtering.Linear.LinearFilters.Gaussian);
            return true;
        }
    }

    class FilterLinearAdaptiveGaussProc : AlgGrayImgProcessor
    {
        FilterAdapGaussParams pars = new FilterAdapGaussParams();

        public override bool Apply(ImageGray img)
        {
            FilterAdapGaussForm form = new FilterAdapGaussForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Linear.AdaptiveGaussFilter(pars.Sigma));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterLinearLowpassProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(ImageSharp.Filtering.Linear.LinearFilters.Lowpass);
            return true;
        }
    }

    class FilterLinearHighpassProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(ImageSharp.Filtering.Linear.LinearFilters.Highpass);
            return true;
        }
    }

    class FilterLinearHighboostProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(ImageSharp.Filtering.Linear.LinearFilters.Highboost);
            return true;
        }
    }

    class FilterLinearLaplacianProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(ImageSharp.Filtering.Linear.LinearFilters.Laplacian);
            return true;
        }
    }

    class FilterLinearLaplacianSharpenProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(ImageSharp.Filtering.Linear.LinearFilters.LaplacianSharpen);
            return true;
        }
    }

    class FilterLinearUserDefinedProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            FilterLinearParams pars = new FilterLinearParams();
            if (img != null)
            {
                img.Apply(new ImageSharp.Filtering.Linear.LinearFilter(pars.FilterMatrix));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterNonLinearMedianProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            FilterNonLinMedianParams pars = new FilterNonLinMedianParams();
            if (img != null)
            {
                img.Apply(new ImageSharp.Filtering.NonLinear.MedianFilter());
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterNonLinearUnsharpenProc : AlgGrayImgProcessor
    {
        FilterNonLinUnsharpParams pars = new FilterNonLinUnsharpParams();

        public override bool Apply(ImageGray img)
        {
            FilterNonLinUnsharpForm form = new FilterNonLinUnsharpForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.NonLinear.UnsharpMasking(pars.Sigma, pars.Intensity));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterNonLinearISEFProc : AlgGrayImgProcessor
    {
        FilterNonLinISEFParams pars = new FilterNonLinISEFParams();

        public override bool Apply(ImageGray img)
        {
            FilterNonLinISEFForm form = new FilterNonLinISEFForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.NonLinear.ISEFFilter(pars.SmoothingFactor));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterNonLinearHighboostProc : AlgGrayImgProcessor
    {
        FilterNonLinHighboostParams pars = new FilterNonLinHighboostParams();

        public override bool Apply(ImageGray img)
        {
            FilterNonLinHighboostForm form = new FilterNonLinHighboostForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.NonLinear.HighboostFilter(pars.Sigma));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterNonLinearHighboost2Proc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.Filtering.NonLinear.HighboostFilter2());
            return true;
        }
    }

    class FilterSpectralLaplaceProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.Filtering.Spectral.LaplacianSpectral());
            return true;
        }
    }

    class FilterSpectralLoGProc : AlgGrayImgProcessor
    {
        FilterSpectralLoGParams pars = new FilterSpectralLoGParams();

        public override bool Apply(ImageGray img)
        {
            FilterSpectralLoGForm form = new FilterSpectralLoGForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Spectral.LaplacianOfGaussianSpectral(pars.Sigma));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterSpectralBandpassProc : AlgGrayImgProcessor
    {
        FilterSpectralBandpassParams pars = new FilterSpectralBandpassParams();

        public override bool Apply(ImageGray img)
        {
            pars.SetRanges(img.Width, img.Height);
            FilterSpectralBandpassForm form = new FilterSpectralBandpassForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Spectral.BandpassSpectral(pars.Min, pars.Max));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterSpectralLowpassProc : AlgGrayImgProcessor
    {
        FilterSpectralLowpassParams pars = new FilterSpectralLowpassParams();

        public override bool Apply(ImageGray img)
        {
            pars.SetRanges(img.Width, img.Height);
            FilterSpectralLowpassForm form = new FilterSpectralLowpassForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Spectral.LowpassSpectral(pars.Cutoff));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterSpectralHighpassProc : AlgGrayImgProcessor
    {
        FilterSpectralHighpassParams pars = new FilterSpectralHighpassParams();
        
        public override bool Apply(ImageGray img)
        {
            pars.SetRanges(img.Width, img.Height);
            FilterSpectralHighpassForm form = new FilterSpectralHighpassForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Spectral.HighpassSpectral(pars.Cutoff));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterSpectralSmoothBandpassProc : AlgGrayImgProcessor
    {
        FilterSpectralSmoothBandpassParams pars = new FilterSpectralSmoothBandpassParams();

        public override bool Apply(ImageGray img)
        {
            pars.SetRanges(img.Width, img.Height);
            FilterSpectralSmoothBandpassForm form = new FilterSpectralSmoothBandpassForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Spectral.SmoothBandpassFilter(pars.Min, pars.Omega1, pars.Max, pars.Omega2));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterSpectralSmoothLowpassProc : AlgGrayImgProcessor
    {
        FilterSpectralSmoothLowpassParams pars = new FilterSpectralSmoothLowpassParams();

        public override bool Apply(ImageGray img)
        {
            pars.SetRanges(img.Width, img.Height);
            FilterSpectralSmoothLowpassForm form = new FilterSpectralSmoothLowpassForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Spectral.SmoothLowpassFilter(pars.Cutoff, pars.Omega));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class FilterSpectralSmoothHighpassProc : AlgGrayImgProcessor
    {
        FilterSpectralSmoothHighpassParams pars = new FilterSpectralSmoothHighpassParams();

        public override bool Apply(ImageGray img)
        {
            pars.SetRanges(img.Width, img.Height);
            FilterSpectralSmoothHighpassForm form = new FilterSpectralSmoothHighpassForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Filtering.Spectral.SmoothHighpassFilter(pars.Cutoff, pars.Omega));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class EdgesCannyProc : AlgGrayImgProcessor
    {
        EdgesCannyParams pars = new EdgesCannyParams();

        public override bool Apply(ImageGray img)
        {
            EdgesCannyForm form = new EdgesCannyForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                byte tLow = pars.ThreshLow;
                byte tHigh = pars.ThreshHigh;
                if (pars.AutoThresh)
                {
                    tLow = 1;
                    tHigh = 0;
                }
                img.Apply(new ImageSharp.Detections.EdgeDetectors.Optimal.CannyDetector(pars.Sigma, tLow, tHigh, pars.GradOnly));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class EdgesShenCastanProc : AlgGrayImgProcessor
    {
        EdgesShenCastanParams pars = new EdgesShenCastanParams();
        public override bool Apply(ImageGray img)
        {
            EdgesShenCastanForm form = new EdgesShenCastanForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Detections.EdgeDetectors.Optimal.ShenCastanDetector(pars.SmoothFac, pars.ThreshLow, pars.ThreshHigh, pars.WinWidth, pars.ThinFactor));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class EdgesMarrHildrethProc : AlgGrayImgProcessor
    {
        EdgesMarrHildrethParams pars = new EdgesMarrHildrethParams();

        public override bool Apply(ImageGray img)
        {
            EdgesMarrHildrethForm form = new EdgesMarrHildrethForm(pars);
            if (form.ShowDialog() == DialogResult.OK)
            {
                img.Apply(new ImageSharp.Detections.EdgeDetectors.Optimal.MarrHildrethDetector((float)pars.Sigma));
                return true;
            }
            else
            {
                return false;
            }
        }
    }

    class EdgesRobertsProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.Detections.EdgeDetectors.Gradient.Roberts());
            return true;
        }
    }

    class EdgesPrewittProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.Detections.EdgeDetectors.Gradient.Prewitt());
            return true;
        }
    }

    class EdgesSobelProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.Detections.EdgeDetectors.Gradient.Sobel());
            return true;
        }
    }

    class EdgesLaplaceProc : AlgGrayImgProcessor
    {
        public override bool Apply(ImageGray img)
        {
            img.Apply(new ImageSharp.Detections.EdgeDetectors.Derivative2.Laplace());
            return true;
        }
    }
}