using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Collections;
using System.Windows.Forms;
using System.Drawing;
using ImageSharp.Maths;

namespace ImageSharp.Camera
{
    #region Hilfsstrukturen und Enumerationen
    /// <summary>
    /// Enumeration fr die Maeinheit der bergebenen Kalibrierungs- sowie Lichtraumprofil-Daten.
    /// </summary>
    enum MeasureUnit
    {
        Millimeter = 0,
        None = Millimeter,
        Centimeter = 1,
        Decimeter = 2,
        Meter = 3,
        Kilometer = 4,
        HighestUnit = Kilometer
    };

    /// <summary>
    /// Skalierungsfaktoren fr die einzelnen Elemente aus MeasureUnit.
    /// </summary>
    sealed class MeasureScales
    {
        public static int Scale(MeasureUnit unit){
            switch (unit)
            {
                case MeasureUnit.Millimeter:
                    return 1;
                case MeasureUnit.Centimeter:
                    return 10;
                case MeasureUnit.Decimeter:
                    return 100;
                case MeasureUnit.Meter:
                    return 1000;
                case MeasureUnit.Kilometer:
                    return 1000000;
            }
            return 1;
        }
    }

    /// <summary>
    /// 2D-Punkt, der die Koordinaten des Bildkoordinatensystems wiedergeben soll.
    /// Vergleichbar mit der Klasse Point.
    /// </summary>
    struct Point2D
    {
        /// <summary>
        /// X-Koordinate des Punktes.
        /// </summary>
        public double X;
        /// <summary>
        /// Y-Koordinate des Punktes.
        /// </summary>
        public double Y;

        public Point2D(double x, double y)
        {
            this.X = x;
            this.Y = y;
        }

        public Point2D(Point2D point)
        {
            this.X = point.X;
            this.Y = point.Y;
        }

        public Point2D(Point point)
        {
            this.X = point.X;
            this.Y = point.Y;
        }

        /// <summary>
        /// Wandelt den Point2D in ein Point-Objekt um.
        /// </summary>
        /// <returns>Zu diesem Point2D gehrendes Point-Objekt.</returns>
        public Point ToPoint()
        {
            return new Point((int)this.X, (int)this.Y);
        }
    }

    /// <summary>
    /// 3D-Punkt zur Aufnahme eines Punktes im Weltkoordinatensystem.
    /// </summary>
    struct Point3D
    {
        /// <summary>
        /// X-Koordinate des Punktes.
        /// </summary>
        public double X;
        /// <summary>
        /// Y-Koordinate des Punktes.
        /// </summary>
        public double Y;
        /// <summary>
        /// Z-Koordinate des Punktes.
        /// </summary>
        public double Z;

        public Point3D(double x, double y, double z)
        {
            this.X = x;
            this.Y = y;
            this.Z = z;
        }

        public Point3D(Point3D point)
        {
            this.X = point.X;
            this.Y = point.Y;
            this.Z = point.Z;
        }

        /// <summary>
        /// Erzeut einen 3D-Punkt aus dem bergebenen Point2D-Objekt, Z wird 0 gesetzt.
        /// </summary>
        /// <param name="point">Punkt, aus dem der 3D-Punkt erstellt werden soll.</param>
        public Point3D(Point2D point)
        {
            this.X = point.X;
            this.Y = point.Y;
            this.Z = 0.0;
        }

        /// <summary>
        /// Erzeut einen 3D-Punkt aus dem bergebenen Point-Objekt, Z wird 0 gesetzt.
        /// </summary>
        /// <param name="point">Punkt, aus dem der 3D-Punkt erstellt werden soll.</param>
        public Point3D(Point point)
        {
            this.X = point.X;
            this.Y = point.Y;
            this.Z = 0.0;
        }
    }

    /// <summary>
    /// Struktur fr einen Passpunkt. Dieser besteht aus einem Welt- und dem zugehrigen Bildpunkt.
    /// </summary>
    struct PassPoint
    {
        /// <summary>
        /// Bildpunkt des Bild-/Weltpunkt Passpunkt-Paares.
        /// </summary>
        public Point2D PointImg;
        /// <summary>
        /// Weltpunkt des Bild-/Weltpunkt Passpunkt-Paares.
        /// </summary>
        public Point3D PointWorld;

        /// <summary>
        /// Erzeugt einen Passpunkt anhand des bergebenen Bild-/Weltpunkt-Paares.
        /// </summary>
        /// <param name="pointImg">Bildpunkt des Passpunktes</param>
        /// <param name="pointWorld">Weltpunkt des Passpunktes</param>
        public PassPoint(Point2D pointImg, Point3D pointWorld)
        {
            this.PointImg = pointImg;
            this.PointWorld = pointWorld;
        }

        public PassPoint(PassPoint pair)
        {
            this.PointImg = new Point2D(pair.PointImg);
            this.PointWorld = new Point3D(pair.PointWorld);
        }
    }
    #endregion
    
    /// <summary>
    /// Grundlegende Photogrammetrie-Klasse. Stellt mathematischen Zusammenhang zwischen
    /// Welt- und Bildkoordinaten her. Bietet Funktionen, mit denen man Punkte vom einen in
    /// das andere Koordinatensystem umrechnen kann nach erfolgter Kalibrierung der Kamera ber
    /// 6 Passpunkte mittels Direkter Linearer Transformation (DLT).
    /// </summary>
    class Calibration
    {
        #region Private Attribute
        private MeasureUnit measureUnit = MeasureUnit.None; // Maeinheit der Kalibrierung
        private double fixedWidth = 640;                    // Normalisierte Videobreite
        private double imgWidth = 640;                      // wird durch die tatschliche Bildbreite berschrieben
        private ArrayList pointsCalib = new ArrayList();    // Speicherstruktur fr Passpunkte

        // DLT-Parameter
        private double[] L = null;

        // Parameter der inneren Orientierung
        private double x0, y0;
        private double fx, fy, f;
        private double s, m;

        // Parameter der ueren Orientierung
        private double Xw0, Yw0, Zw0;
        private double[,] R = new double[3, 3];
        private double omega, phi, kappa;

        // XML-Tag- und -Attribut-Namen
        private String XmlCalibElemRoot = "ImageSharpCalibration";
        private String XmlCalibElemMeasureData = "measureData";
        private String XmlCalibElemCalibrations = "calibrations";
        private String XmlCalibElemCalibration = "calibration";
        private String XmlCalibElemCalibData = "calibData";
        private String XmlCalibAttribCalibDataUnit = "unit";
        private String XmlCalibElemPoint = "point";
        private String XmlCalibElemPointWorld = "world";
        private String XmlCalibElemPointImg = "img";
        private String XmlCalibAttribPointX = "x";
        private String XmlCalibAttribPointY = "y";
        private String XmlCalibAttribPointZ = "z";
        #endregion

        #region ffentliche Eigenschaften und Methoden
        /// <summary>
        /// Gibt die Maeinheit der Kalibrierung zurck.
        /// </summary>
        public MeasureUnit Unit
        {
            get
            {
                return measureUnit;
            }
        }

        /// <summary>
        /// Erstellen einer Kalibrierung anhand der bergebenen XML-Datei.
        /// Diese enthlt mind. 6 Passpunkte.
        /// </summary>
        /// <param name="xmlFile">Auszulesende XML-Datei</param>
        /// <param name="imgWidth">Tatschliche Breite des Videobildes, erforderlich zur Umrechnung
        ///     der normalisierten Bildkoordinaten in tatschliche Bildkoordinaten</param>
        public Calibration(string xmlFile, double imgWidth)
        {
            this.imgWidth = imgWidth;
            XmlDocument xmlDoc = new XmlDocument();
            try
            {
                xmlDoc.Load(xmlFile);
            }
            catch
            {
                throw new Exception("Kalibrierungs-Datei konnte nicht geladen werden.");
            }
            readCalibFromXmlDocument(xmlDoc);
            calibrate();
        }

        public Calibration(XmlDocument doc, double imgWidth)
        {
            this.imgWidth = imgWidth;
            readCalibFromXmlDocument(doc);
            calibrate();
        }

        /// <summary>
        /// Erstellt eine Kalibrierung anhand der bergebenen (mind. 6) Passpunkte.
        /// </summary>
        /// <param name="pointPairs">Passpunkte-Array</param>
        /// <param name="imgWidth">Breite der auszuwertenden Bilder</param>
        public Calibration(PassPoint[] pointPairs, double imgWidth)
        {
            this.imgWidth = imgWidth;
            if (pointPairs.Length < 6)
            {
                throw new Exception("Fr die Kalibrierung werden mindestens 6 Punkte-Paare bentigt.");
            }

            foreach (PassPoint pair in pointPairs)
            {
                this.pointsCalib.Add(new PassPoint(pair));
            }

            calibrate();
        }

        /// <summary>
        /// Wandelt Welt- in Bildkoordinaten um.
        /// </summary>
        /// <param name="p">3D-Weltpunkt, der in Bildkoordinaten umgerechnet werden soll.</param>
        /// <returns>Mit p im Bild korrespondierender Bildpunkt.</returns>
        public Point2D WorldToImageCoords(Point3D p)
        {
            Point2D point = new Point2D();

            // Bildpunkt aus den DLT-Parametern ermitteln
            point.X = (L[0] * p.X + L[1] * p.Y + L[2] * p.Z + L[3]) / (L[8] * p.X + L[9] * p.Y + L[10] * p.Z + 1);
            point.Y = (L[4] * p.X + L[5] * p.Y + L[6] * p.Z + L[7]) / (L[8] * p.X + L[9] * p.Y + L[10] * p.Z + 1);

            // Skalierungsfaktor aus den normalisierten Punkten berechnen
            double scaleFac = imgWidth / fixedWidth;
            point.X *= scaleFac;
            point.Y *= scaleFac;

            return point;
        }

        /// <summary>
        /// Wandelt den bergebenen Bildpunkt in Weltkoordinaten um. Dabei ist es erforderlich,
        /// in die Z=0-Ebene hinein zu transformieren, da jeder Punkt im Bildkoordinatensystem mit einem
        /// ganzen Strahl von Punkten im Weltkoordinatensystem korrespondiert.
        /// </summary>
        /// <param name="p">Umzuwandelnder 2D-Bildpunkt</param>
        /// <returns>Mit p korrespondierender 3D-Weltpunkt in der Z=0-Ebene</returns>
        public Point3D ImageToRailLayerCoords(Point2D p)
        {
            // Skalierungsfaktor aus den normalisierten Punkten berechnen
            double scaleFac = fixedWidth / imgWidth;
            Point2D p1 = new Point2D(p);
            p1.X *= scaleFac;
            p1.Y *= scaleFac;
            Point3D point = new Point3D();

            // Lineares Gleichungssystem lsen, welches sich aus dem skalierten Bildpunkt 
            // und den DLT-Parametern ergibt.
            double[,] A = {
                { L[0] - p1.X*L[8], L[1] - p1.X*L[9] },
                { L[4] - p1.Y*L[8], L[5] - p1.Y*L[9] } };
            double[] b = { p1.X - L[3], p1.Y - L[7] };
            double[] x = MathCalc.SolveLinEqnSystem(A, b);

            // Lsung des LGS gibt den Weltpunkt an.
            point.X = x[0];
            point.Y = x[1];
            point.Z = 0;

            return point;
        }
        #endregion

        #region Private Methoden
        // Liest die Kalibrierung aus dem bergebenen XmlDocument aus.
        private void readCalibFromXmlDocument(XmlDocument xmlDoc)
        {
            // Alte Passpunkte lschen.
            this.pointsCalib.Clear();
            // Bis zu dem Kalibrierungs-Tag vorstoen...
            XmlNodeList nodesRoot = xmlDoc.GetElementsByTagName(XmlCalibElemRoot);
            if (nodesRoot.Count > 0)
            {
                XmlNodeList nodesMeasureData = (nodesRoot[0] as XmlElement).GetElementsByTagName(XmlCalibElemMeasureData);
                if (nodesMeasureData.Count > 0)
                {
                    XmlNodeList nodesCalibrations = (nodesMeasureData[0] as XmlElement).GetElementsByTagName(XmlCalibElemCalibrations);
                    if (nodesCalibrations.Count > 0)
                    {
                        XmlNodeList nodesCalibration = (nodesCalibrations[0] as XmlElement).GetElementsByTagName(XmlCalibElemCalibration);
                        if (nodesCalibration.Count > 0)
                        {
                            XmlNodeList nodesCalibData = (nodesCalibration[0] as XmlElement).GetElementsByTagName(XmlCalibElemCalibData);
                            if (nodesCalibData.Count > 0)
                            {
                                // hier sind die eigentlichen Passpunkte enthalten.
                                // Das Attribut "unit" kann zustzlich die Maeinheit angeben.
                                XmlElement elemCalibData = nodesCalibData[0] as XmlElement;
                                XmlAttribute attribUnit = elemCalibData.GetAttributeNode(XmlCalibAttribCalibDataUnit);
                                if (attribUnit != null)
                                {
                                    this.measureUnit = (MeasureUnit)Int32.Parse(attribUnit.Value);
                                }
                                double scale = (double)MeasureScales.Scale(measureUnit);

                                // Alle vorhandenen Passpunkte auslesen.
                                XmlNodeList nodesPoints = elemCalibData.GetElementsByTagName(XmlCalibElemPoint);
                                foreach (XmlElement elemPoint in nodesPoints)
                                {
                                    XmlNodeList nodesWorld = elemPoint.GetElementsByTagName(XmlCalibElemPointWorld);
                                    XmlNodeList nodesImage = elemPoint.GetElementsByTagName(XmlCalibElemPointImg);

                                    if (nodesWorld.Count == 0)
                                        throw new Exception("Zu einem Kalibrierungs-Punkt sind keine Weltkoordination angegeben.");
                                    else if (nodesImage.Count == 0)
                                        throw new Exception("Zu einem Kalibrierungs-Punkt sind keine Bildkoordination angegeben.");

                                    XmlElement elemWorld = nodesWorld[0] as XmlElement;
                                    XmlElement elemImage = nodesImage[0] as XmlElement;

                                    // Attribute eines Punktes lesen - geben die Koordinaten des jeweiligen
                                    // Bild- bzw. Weltpunktes an.
                                    XmlAttribute attribWorldX = elemWorld.GetAttributeNode(XmlCalibAttribPointX);
                                    XmlAttribute attribWorldY = elemWorld.GetAttributeNode(XmlCalibAttribPointY);
                                    XmlAttribute attribWorldZ = elemWorld.GetAttributeNode(XmlCalibAttribPointZ);
                                    XmlAttribute attribImageX = elemImage.GetAttributeNode(XmlCalibAttribPointX);
                                    XmlAttribute attribImageY = elemImage.GetAttributeNode(XmlCalibAttribPointY);

                                    if (attribWorldX == null)
                                        throw new Exception("Zu einem Kalibrierungs-Punkt im Weltkoordinatensystem ist keine x-Koordinate angegeben.");
                                    else if (attribWorldY == null)
                                        throw new Exception("Zu einem Kalibrierungs-Punkt im Weltkoordinatensystem ist keine y-Koordinate angegeben.");
                                    else if (attribWorldZ == null)
                                        throw new Exception("Zu einem Kalibrierungs-Punkt im Weltkoordinatensystem ist keine z-Koordinate angegeben.");
                                    else if (attribImageX == null)
                                        throw new Exception("Zu einem Kalibrierungs-Punkt im Bildkoordinatensystem ist keine x-Koordinate angegeben.");
                                    else if (attribImageY == null)
                                        throw new Exception("Zu einem Kalibrierungs-Punkt im Bildkoordinatensystem ist keine y-Koordinate angegeben.");

                                    // Wandle die Koordinaten-Attribute in double-Zahlen um.
                                    double worldX = MathCalc.ParseDouble(attribWorldX.Value);
                                    double worldY = MathCalc.ParseDouble(attribWorldY.Value);
                                    double worldZ = MathCalc.ParseDouble(attribWorldZ.Value);
                                    double imageX = MathCalc.ParseDouble(attribImageX.Value);
                                    double imageY = MathCalc.ParseDouble(attribImageY.Value);
                                    if (worldX == Double.NaN || worldY == Double.NaN || worldZ == Double.NaN
                                       || imageX == Double.NaN || imageY == Double.NaN)
                                        throw new Exception("Eine ausgelesene Punktkoordinate konnte nicht in double umgewandelt werden.");

                                    // Ausgelesene Punkte auf Millimeter skalieren und neuen Passpunkt hinzufgen
                                    this.pointsCalib.Add(new PassPoint(new Point2D(imageX, imageY), new Point3D(worldX * scale, worldY * scale, worldZ * scale)));
                                }

                                if (this.pointsCalib.Count < 6)
                                {
                                    throw new Exception("Fr die Kalibrierung werden mindestens 6 Punkte-Paare bentigt.");
                                }
                            }
                            else
                            {
                                throw new Exception("Kalibrierung konnte nicht gelesen werden (kein Daten-Element innerhalb einer Kalibrierung)");
                            }
                        }
                        else
                        {
                            throw new Exception("Kalibrierung konnte nicht gelesen werden (keine Kalibrierung gefunden)");
                        }
                    }
                    else
                    {
                        throw new Exception("Kalibrierung konnte nicht gelesen werden (kein gltiges Kalibrierungen-Element)");
                    }
                }
                else
                {
                    throw new Exception("Kalibrierung konnte nicht gelesen werden (kein gltiges Messdatenelement)");
                }
            }
            else
            {
                throw new Exception("Kalibrierung konnte nicht gelesen werden (kein gltiges Wurzelelement)");
            }
        }

        // Erstellt mit den vorhandenen Passpunkten die Kalibrierung.
        private void calibrate()
        {
            // Array fr das lineare Gleichungssystem der DLT.
            double[,] A = new double[2 * this.pointsCalib.Count, 11];
            double[] b = new double[2 * this.pointsCalib.Count];

            // Erstelle die Gleichungssystem-Arrays.
            int i = 0;
            foreach (PassPoint point in pointsCalib)
            {
                A[i, 0] = point.PointWorld.X;
                A[i, 1] = point.PointWorld.Y;
                A[i, 2] = point.PointWorld.Z;
                A[i, 3] = 1;
                A[i, 8] = -point.PointImg.X * point.PointWorld.X;
                A[i, 9] = -point.PointImg.X * point.PointWorld.Y;
                A[i, 10] = -point.PointImg.X * point.PointWorld.Z;

                A[i + 1, 4] = point.PointWorld.X;
                A[i + 1, 5] = point.PointWorld.Y;
                A[i + 1, 6] = point.PointWorld.Z;
                A[i + 1, 7] = 1;
                A[i + 1, 8] = -point.PointImg.Y * point.PointWorld.X;
                A[i + 1, 9] = -point.PointImg.Y * point.PointWorld.Y;
                A[i + 1, 10] = -point.PointImg.Y * point.PointWorld.Z;

                b[i] = point.PointImg.X;
                b[i + 1] = point.PointImg.Y;

                i += 2;
            }
            
            // Lse das DLT-Gleichungssystem.
            this.L = MathCalc.SolveLinEqnSystem(A, b);
            
            // Speichere die Parameter der inneren und ueren Orientierung durch Rcksubstitution.
            double Lquad = L[8] * L[8] + L[9] * L[9] + L[10] * L[10];
            double Lvar = Math.Sqrt(Lquad);
            double aTc = L[0] * L[8] + L[1] * L[9] + L[2] * L[10];
            double bTc = L[4] * L[8] + L[5] * L[9] + L[6] * L[10];
            double aTb = L[0] * L[4] + L[1] * L[5] + L[2] * L[6];
            double aTa = L[0] * L[0] + L[1] * L[1] + L[2] * L[2];
            double bTb = L[4] * L[4] + L[5] * L[5] + L[6] * L[6];
            double[,] abc = { { L[0], L[1], L[2] }, { L[4], L[5], L[6] }, { L[8], L[9], L[10] } };
            double abcDet = abc[0, 0] * abc[1, 1] * abc[2, 2] + abc[0, 1] * abc[1, 2] * abc[2, 0] + abc[0, 2] * abc[1, 0] * abc[2, 1]
                - abc[0, 0] * abc[1, 2] * abc[2, 1] - abc[0, 1] * abc[1, 0] * abc[2, 2] - abc[0, 2] * abc[1, 1] * abc[2, 0];
            double[,] abcInv = { 
                {(abc[1,1]*abc[2,2] - abc[1,2]*abc[2,1])/abcDet, (abc[0,2]*abc[2,1] - abc[0,1]*abc[2,2])/abcDet, (abc[0,1]*abc[1,2] - abc[0,2]*abc[1,1])/abcDet}, 
                {(abc[1,2]*abc[2,0] - abc[1,0]*abc[2,2])/abcDet, (abc[0,0]*abc[2,2] - abc[0,2]*abc[2,0])/abcDet, (abc[0,2]*abc[1,0] - abc[0,0]*abc[1,2])/abcDet},
                {(abc[1,0]*abc[2,1] - abc[1,1]*abc[2,0])/abcDet, (abc[0,1]*abc[2,0] - abc[0,0]*abc[2,1])/abcDet, (abc[0,0]*abc[1,1] - abc[0,1]*abc[1,0])/abcDet} };

            this.x0 = aTc / Lquad;
            this.y0 = bTc / Lquad;
            this.fx = Math.Sqrt(aTa / Lquad - this.x0 * this.x0);
            this.fy = Math.Sqrt(bTb / Lquad - this.y0 * this.y0);
            this.f = (this.fx + this.fy) / 2;
            this.s = (aTb * Lquad - aTc * bTc) / (aTa * Lquad - aTc * aTc);
            this.m = -abcDet / (Lvar * Lvar * Lvar * this.fx * this.fx);
            this.Xw0 = -L[3] * abcInv[0, 0] - L[7] * abcInv[0, 1] - abcInv[0, 2];
            this.Yw0 = -L[3] * abcInv[1, 0] - L[7] * abcInv[1, 1] - abcInv[1, 2];
            this.Zw0 = -L[3] * abcInv[2, 0] - L[7] * abcInv[2, 1] - abcInv[2, 2];
            this.R[0, 2] = L[8] / Lvar;
            this.R[1, 2] = L[9] / Lvar;
            this.R[2, 2] = L[10] / Lvar;
            this.R[0, 0] = (this.x0 * this.R[0, 2] - L[0] / Lvar) / this.fx;
            this.R[1, 0] = (this.x0 * this.R[1, 2] - L[1] / Lvar) / this.fx;
            this.R[2, 0] = (this.x0 * this.R[2, 2] - L[2] / Lvar) / this.fx;
            this.R[0, 1] = (this.y0 * this.R[0, 2] - L[4] / Lvar) / this.fy;
            this.R[1, 1] = (this.y0 * this.R[1, 2] - L[5] / Lvar) / this.fy;
            this.R[2, 1] = (this.y0 * this.R[2, 2] - L[6] / Lvar) / this.fy;
            this.omega = Math.Atan(-this.R[1, 2] / this.R[2, 2]);
            this.phi = Math.Asin(-Math.Sign(this.m) * this.R[0, 2]);
            this.kappa = Math.Atan(-this.R[0, 1] / this.R[0, 0]);
        }
        #endregion
    }
}
