using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;
using ImageSharp.Filtering.Linear;
using System.Collections;

namespace ImageSharp
{
    class Geometrics
    {
        // static
    }

    class SeamCarving : IGrayImgApplicable, IColorImgApplicable
    {
        private enum SeamDirection
        {
            Horizontal,
            Vertical
        }

        private class Seam : IComparable<Seam>, IComparable
        {
            public SeamDirection direction = SeamDirection.Horizontal;
            public int[] seamArray = null;
            public float seamValue = 0;

            public int[] SeamArray
            {
                get
                {
                    return this.seamArray;
                }
            }

            public float SeamValue
            {
                get
                {
                    return this.seamValue;
                }
                set
                {
                    this.seamValue = value;
                }
            }

            public SeamDirection Direction
            {
                get
                {
                    return this.direction;
                }
            }

            public Seam(int length, SeamDirection dir)
            {
                this.seamArray = new int[length];
                this.direction = dir;
            }

            public int CompareTo(Seam other)
            {
                return this.SeamValue.CompareTo(other.SeamValue);
            }

            public int CompareTo(object obj)
            {
                if (obj is Seam)
                    return this.CompareTo((Seam)obj);
                else
                    throw new Exception("The method or operation is not implemented.");
            }
        }

        private struct ImgPair
        {
            public RGBData[,] ColorData;
            public byte[,] GrayData;
            public ImgPair(RGBData[,] color, byte[,] gray)
            {
                this.ColorData = color;
                this.GrayData = gray;
            }
        }

        private struct SeamPair
        {
            public Seam[] Seams;
            public bool[,] Reserved;
            public SeamPair(Seam[] seams, bool[,] reserved)
            {
                this.Seams = seams;
                this.Reserved = reserved;
            }
        }

        private struct IntPair : IComparable<IntPair>
        {
            public int Index;
            public int Value;
            public IntPair(int index, int value)
            {
                this.Index = index;
                this.Value = value;
            }
            
            public int CompareTo(IntPair other)
            {
                return this.Value.CompareTo(other.Value);
            }
        }
        
        private int nWidth;
        private int nHeight;

        public SeamCarving(int nWidth, int nHeight)
        {
            this.nWidth = nWidth;
            this.nHeight = nHeight;
        }

        public void Apply(ImageGray img)
        {
            ImageColor nimg = new ImageColor(img);
            rescale(nimg, img);
            img.CreateFrom(nimg.Width, nimg.Height);
            for (int y = 0; y < img.Height; y++)
                for (int x = 0; x < img.Width; x++)
                    img.Data[y, x] = nimg.Data[y, x].R;
        }

        public void Apply(ImageColor img)
        {
            rescale(img, new ImageGray(img));
        }

        private void rescale(ImageColor img, ImageGray grad)
        {
            //grad.Apply(new ImageSharp.Filtering.NonLinear.ISEFFilter(0.5f));
            grad.Apply(GradientFilters.Sobel);

            RGBData[,] data = img.Data;
            byte[,] gradient = grad.Data;

            SeamDirection curDir = (nHeight < data.GetLength(0)) ? SeamDirection.Horizontal : SeamDirection.Vertical;
            while (nWidth < data.GetLength(1) || nHeight < data.GetLength(0))
            {
                Seam seam = findBestSeam(data, gradient, curDir);
                ImgPair p = excludeSeamFromImg(seam, data, gradient);
                data = p.ColorData;
                gradient = p.GrayData;

                if (curDir == SeamDirection.Horizontal)
                {
                    if (nWidth < data.GetLength(1))
                        curDir = SeamDirection.Vertical;
                }
                else
                {
                    if (nHeight < data.GetLength(0))
                        curDir = SeamDirection.Horizontal;
                }
            }

            /*// shrink horizontally with vertical seams
            while (nWidth < data.GetLength(1))
            {
                SeamPair s = findBestSeams(data, gradient, SeamDirection.Vertical, data.GetLength(1) - nWidth);
                ImgPair p = excludeSeamsFromImg(s, SeamDirection.Vertical, data, gradient);
                data = p.ColorData;
                gradient = p.GrayData;
            }

            // shrink vertically with horizontal seams
            while (nHeight < data.GetLength(0))
            {
                SeamPair s = findBestSeams(data, gradient, SeamDirection.Horizontal, data.GetLength(0) - nHeight);
                ImgPair p = excludeSeamsFromImg(s, SeamDirection.Horizontal, data, gradient);
                data = p.ColorData;
                gradient = p.GrayData;
            }*/


            // extend horizontally with vertical seams
            while (nWidth > data.GetLength(1))
            {
                SeamPair s = findBestSeams(data, gradient, SeamDirection.Vertical, nWidth - data.GetLength(1));
                ImgPair p = includeSeamsToImg(s, data, gradient, SeamDirection.Vertical);
                data = p.ColorData;
                gradient = p.GrayData;
            }
            // extend vertically with horizontal seams
            while (nHeight > data.GetLength(0))
            {
                SeamPair s = findBestSeams(data, gradient, SeamDirection.Horizontal, nHeight - data.GetLength(0));
                ImgPair p = includeSeamsToImg(s, data, gradient, SeamDirection.Horizontal);
                data = p.ColorData;
                gradient = p.GrayData;
            }

            img.CopyFrom(data);
        }

        private Seam findBestSeam(RGBData[,] img, byte[,] grad, SeamDirection dir)
        {
            Seam bestSeam = null;
            bool horiz = (dir == SeamDirection.Horizontal);
            int height = img.GetLength(0);
            int width = img.GetLength(1);
            int bound = horiz ? height : width;
            int[,] M = new int[height, width];
            
            if (horiz)
            {
                for (int i = 0; i < height; i++)
                    M[i, width-1] = grad[i, width-1];
                for (int j = width - 2; j >= 0; j--)
                {
                    for (int i = 0; i < height; i++)
                    {
                        int m1, m2, m3;
                        m2 = M[i, j + 1];
                        m1 = m3 = Int32.MaxValue;
                        if (i > 0)
                            m1 = M[i - 1, j + 1];
                        if (i < height - 1)
                            m3 = M[i + 1, j + 1];
                        M[i, j] = grad[i, j] + Math.Min(m1, Math.Min(m2, m3));
                    }
                }

                IntPair[] seamStarts = new IntPair[height];
                for (int i = 0; i < height; i++)
                    seamStarts[i] = new IntPair(i, M[i, 0]);
                Array.Sort<IntPair>(seamStarts);
                int i0 = seamStarts[0].Index;
                bestSeam = new Seam(width, SeamDirection.Horizontal);
                bestSeam.SeamArray[0] = i0;
                bestSeam.SeamValue = seamStarts[0].Value;
                for (int j = 1; j < width; j++)
                {
                    int i = bestSeam.SeamArray[j - 1];
                    if (bestSeam.SeamArray[j - 1] > 0)
                    {
                        int i1 = bestSeam.SeamArray[j - 1] - 1;
                        if (M[i1, j] < M[i, j])
                            i = i1;
                    }
                    if (bestSeam.SeamArray[j - 1] < height - 1)
                    {
                        int i2 = bestSeam.SeamArray[j - 1] + 1;
                        if (M[i2, j] < M[i, j])
                            i = i2;
                    }
                    bestSeam.SeamArray[j] = i;
                }
            }
            else
            {
                for (int i = 0; i < width; i++)
                    M[height-1, i] = grad[height-1, i];
                for (int i = height - 2; i >= 0; i--)
                {
                    for (int j = 0; j < width; j++)
                    {
                        int m1, m2, m3;
                        m2 = M[i + 1, j];
                        m1 = m3 = Int32.MaxValue;
                        if (j > 0)
                            m1 = M[i + 1, j - 1];
                        if (j < width - 1)
                            m3 = M[i + 1, j + 1];
                        M[i, j] = grad[i, j] + Math.Min(m1, Math.Min(m2, m3));
                    }
                }

                IntPair[] seamStarts = new IntPair[width];
                for (int i = 0; i < width; i++)
                    seamStarts[i] = new IntPair(i, M[0, i]);
                Array.Sort<IntPair>(seamStarts);
                int i0 = seamStarts[0].Index;
                bestSeam = new Seam(height, SeamDirection.Vertical);
                bestSeam.SeamArray[0] = i0;
                bestSeam.SeamValue = seamStarts[0].Value;
                for (int i = 1; i < height; i++)
                {
                    int j = bestSeam.SeamArray[i - 1];
                    if (bestSeam.SeamArray[i - 1] > 0)
                    {
                        int j1 = bestSeam.SeamArray[i - 1] - 1;
                        if (M[i, j1] < M[i, j])
                            j = j1;
                    }
                    if (bestSeam.SeamArray[i - 1] < width - 1)
                    {
                        int j2 = bestSeam.SeamArray[i - 1] + 1;
                        if (M[i, j2] < M[i, j])
                            j = j2;
                    }
                    bestSeam.SeamArray[i] = j;
                }
            }

            return bestSeam;/*

            for (int i = 0; i < bound; i++)
            {
                Seam curSeam = new Seam(horiz ? width : height, dir);
                curSeam.SeamArray[0] = i;
                bool broken = false;
                for (int j = 1; j < curSeam.SeamArray.Length; j++)
                {
                    // search for next seam pixel in current row/col
                    int nextIndex = curSeam.SeamArray[j - 1];
                    int lastSum = 0;
                    int curSum = 0;
                    if (horiz)
                    {
                        lastSum = gradSum[curSeam.SeamArray[j - 1], j - 1];
                        curSum = lastSum + grad[nextIndex, j] * grad[nextIndex, j];
                        if (curSum < gradSum[nextIndex, j])
                            gradSum[nextIndex, j] = curSum;
                    }
                    else
                    {
                        lastSum = gradSum[j - 1, curSeam.SeamArray[j - 1]];
                        curSum = lastSum + grad[j, nextIndex] * grad[j, nextIndex];
                        if (curSum < gradSum[j, nextIndex])
                            gradSum[j, nextIndex] = curSum;
                    }
                    bool hasPredec = nextIndex > 0;         // pixel has a predecessor
                    bool hasDec = nextIndex < bound - 1;    // pixel has a decessor
                    if (hasPredec)
                    {
                        int predec = curSeam.SeamArray[j - 1] - 1;
                        int sum = 0;
                        if (horiz)
                        {
                            sum = lastSum + grad[predec, j] * grad[predec, j];
                            if (sum < gradSum[predec, j])
                                gradSum[predec, j] = sum;
                        }
                        else
                        {
                            sum = lastSum + grad[j, predec] * grad[j, predec];
                            if (sum < gradSum[j, predec])
                                gradSum[j, predec] = sum;
                        }
                        if (sum < curSum)
                        {
                            nextIndex = predec;
                            curSum = sum;
                        }
                    }
                    if (hasDec)
                    {
                        int dec = curSeam.SeamArray[j - 1] + 1;
                        int sum = 0;
                        if (horiz)
                        {
                            sum = lastSum + grad[dec, j] * grad[dec, j];
                            if (sum < gradSum[dec, j])
                                gradSum[dec, j] = sum;
                        }
                        else
                        {
                            sum = lastSum + grad[j, dec] * grad[j, dec];
                            if (sum < gradSum[j, dec])
                                gradSum[j, dec] = sum;
                        }
                        if (sum < curSum)
                        {
                            nextIndex = dec;
                            curSum = sum;
                        }
                    }

                    // now we found an index for minimal gradient sum, so check it to the underlying sum
                    if (curSum > (horiz ? gradSum[nextIndex, j] : gradSum[j, nextIndex]))
                    {
                        broken = true;
                        break;
                    }
                    else
                    {
                        curSeam.SeamArray[j] = nextIndex;
                        curSeam.SeamValue = curSum;
                    }
                }
                if (!broken)
                {
                    if (bestSeam == null || curSeam.SeamValue < bestSeam.SeamValue)
                        bestSeam = curSeam;
                }
            }

            return bestSeam;*/
        }

        /*private Seam findBestSeam(RGBData[,] img, byte[,] grad, SeamDirection dir)
        {
            Seam bestSeam = null;
            bool horiz = (dir == SeamDirection.Horizontal);
            int height = img.GetLength(0);
            int width = img.GetLength(1);
            int bound = horiz ? height : width;
            int[,] gradSum = new int[height, width];
            for (int y = 0; y < height; y++)
                for (int x = 0; x < width; x++)
                    gradSum[y, x] = Int32.MaxValue;
            if(horiz){
                for(int i=0; i<height; i++)
                    gradSum[i, 0] = grad[i, 0];
            }
            else{
                for(int i=0; i<width; i++)
                    gradSum[0, i] = grad[0, i];
            }
            
            for (int i = 0; i < bound; i++)
            {
                Seam curSeam = new Seam(horiz ? width : height, dir);
                curSeam.SeamArray[0] = i;
                bool broken = false;
                for (int j = 1; j < curSeam.SeamArray.Length; j++)
                {
                    // search for next seam pixel in current row/col
                    int nextIndex = curSeam.SeamArray[j - 1];
                    int lastSum = 0;
                    int curSum = 0;
                    if(horiz){
                        lastSum = gradSum[curSeam.SeamArray[j - 1], j - 1];
                        curSum = lastSum + grad[nextIndex, j] * grad[nextIndex, j];
                        if (curSum < gradSum[nextIndex, j])
                            gradSum[nextIndex, j] = curSum;
                    }
                    else{
                        lastSum = gradSum[j - 1, curSeam.SeamArray[j - 1]];
                        curSum = lastSum + grad[j, nextIndex] * grad[j, nextIndex];
                        if (curSum < gradSum[j, nextIndex])
                            gradSum[j, nextIndex] = curSum;
                    }
                    bool hasPredec = nextIndex > 0;         // pixel has a predecessor
                    bool hasDec = nextIndex < bound - 1;    // pixel has a decessor
                    if (hasPredec)
                    {
                        int predec = curSeam.SeamArray[j - 1] - 1;
                        int sum = 0;
                        if(horiz){
                            sum = lastSum + grad[predec, j] * grad[predec, j];
                            if(sum < gradSum[predec, j])
                                gradSum[predec, j] = sum;
                        }
                        else{
                            sum = lastSum + grad[j, predec] * grad[j, predec];
                            if (sum < gradSum[j, predec])
                                gradSum[j, predec] = sum;
                        }
                        if (sum < curSum)
                        {
                            nextIndex = predec;
                            curSum = sum;
                        }
                    }
                    if (hasDec)
                    {
                        int dec = curSeam.SeamArray[j - 1] + 1;
                        int sum = 0;
                        if (horiz)
                        {
                            sum = lastSum + grad[dec, j] * grad[dec, j];
                            if (sum < gradSum[dec, j])
                                gradSum[dec, j] = sum;
                        }
                        else
                        {
                            sum = lastSum + grad[j, dec] * grad[j, dec];
                            if (sum < gradSum[j, dec])
                                gradSum[j, dec] = sum;
                        }
                        if (sum < curSum)
                        {
                            nextIndex = dec;
                            curSum = sum;
                        }
                    }

                    // now we found an index for minimal gradient sum, so check it to the underlying sum
                    if (curSum > (horiz ? gradSum[nextIndex, j] : gradSum[j, nextIndex]))
                    {
                        broken = true;
                        break;
                    }
                    else
                    {
                        curSeam.SeamArray[j] = nextIndex;
                        curSeam.SeamValue = curSum;
                    }
                }
                if (!broken)
                {
                    if (bestSeam == null || curSeam.SeamValue < bestSeam.SeamValue)
                        bestSeam = curSeam;
                }
            }

            return bestSeam;
        }*/

        private ImgPair excludeSeamFromImg(Seam seam, RGBData[,] img, byte[,] grad)
        {
            int height = img.GetLength(0);
            int width = img.GetLength(1);
            int nheight = (seam.Direction == SeamDirection.Horizontal) ? height - 1 : height;
            int nwidth = (seam.Direction == SeamDirection.Horizontal) ? width : width - 1;
            RGBData[,] nimg = new RGBData[nheight, nwidth];
            byte[,] ngrad = new byte[nheight, nwidth];
            
            int bound1 = (seam.Direction == SeamDirection.Horizontal) ? width : height;
            int bound2 = (seam.Direction == SeamDirection.Horizontal) ? height : width;

            for (int i = 0; i < bound1; i++)
            {
                for (int j = 0, k = 0; j < bound2; j++, k++)
                {
                    if(j == seam.SeamArray[i])
                        j++;
                    if (j < bound2)
                    {
                        if (seam.Direction == SeamDirection.Horizontal)
                        {
                            nimg[k, i] = img[j, i];
                            ngrad[k, i] = grad[j, i];
                        }
                        else
                        {
                            nimg[i, k] = img[i, j];
                            ngrad[i, k] = grad[i, j];
                        }
                    }
                }
            }

            return new ImgPair(nimg, ngrad);
        }

        private SeamPair findBestSeams(RGBData[,] img, byte[,] grad, SeamDirection dir, int count)
        {
            int height = img.GetLength(0);
            int width = img.GetLength(1);
            ArrayList seams = new ArrayList();
            bool horiz = (dir == SeamDirection.Horizontal);
            int bound = horiz ? height : width;
            bool[,] reserved = new bool[height, width];

            for (int i = 0; i < bound; i++)
            {
                Seam curSeam = new Seam(horiz ? width : height, dir);
                curSeam.SeamArray[0] = i;
                curSeam.SeamValue = horiz ? grad[i, 0] : grad[0, i];
                bool broken = false;
                for (int j = 1; j < curSeam.SeamArray.Length; j++)
                {
                    // search for next seam pixel in current row/col
                    int nextIndex = curSeam.SeamArray[j - 1];
                    int curgrad = horiz ? grad[nextIndex, j] : grad[j, nextIndex];
                    bool hasPredec = nextIndex > 0;         // pixel has a predecessor
                    bool hasDec = nextIndex < bound - 1;    // pixel has a decessor
                    if (hasPredec)
                    {
                        int predec = curSeam.SeamArray[j - 1] - 1;
                        if (horiz ? !reserved[predec, j] : !reserved[j, predec])
                        {
                            int pregrad = horiz ? grad[predec, j] : grad[j, predec];
                            if (pregrad < curgrad || (horiz ? reserved[nextIndex, j] : reserved[j, nextIndex]))
                            {
                                nextIndex = predec;
                                curgrad = pregrad;
                            }
                        }
                    }
                    if (hasDec)
                    {
                        int dec = curSeam.SeamArray[j - 1] + 1;
                        if (horiz ? !reserved[dec, j] : !reserved[j, dec])
                        {
                            int decgrad = horiz ? grad[dec, j] : grad[j, dec];
                            if (decgrad < curgrad || (horiz ? reserved[nextIndex, j] : reserved[j, nextIndex]))
                            {
                                nextIndex = dec;
                                curgrad = decgrad;
                            }
                        }
                    }

                    // break if the index is reserved for another seam
                    if (horiz ? reserved[nextIndex, j] : reserved[j, nextIndex])
                    {
                        broken = true;
                        break;
                    }
                    else
                    {
                        curSeam.SeamArray[j] = nextIndex;
                        curSeam.SeamValue += curgrad;
                    }
                }
                if (!broken)
                {
                    seams.Add(curSeam);
                    for (int j = 0; j < curSeam.SeamArray.Length; j++)
                    {
                        if (horiz)
                            reserved[curSeam.SeamArray[j], j] = true;
                        else
                            reserved[j, curSeam.SeamArray[j]] = true;
                    }
                }
            }

            seams.Sort();

            Seam[] bestSeams = new Seam[Math.Min(count, seams.Count)];
            int k=0;
            foreach (Seam s in seams)
            {
                bestSeams[k++] = s;
                if (k >= bestSeams.Length)
                    break;
            }

            reserved = new bool[height, width];
            for (k = 0; k < bestSeams.Length; k++)
            {
                for (int j = 0; j < bestSeams[k].SeamArray.Length; j++)
                {
                    if (horiz)
                        reserved[bestSeams[k].SeamArray[j], j] = true;
                    else
                        reserved[j, bestSeams[k].SeamArray[j]] = true;
                }
            }

            return new SeamPair(bestSeams, reserved);
        }

        private ImgPair includeSeamsToImg(SeamPair s, RGBData[,] img, byte[,] grad, SeamDirection dir)
        {
            int height = img.GetLength(0);
            int width = img.GetLength(1);
            Seam[] seams = s.Seams;
            bool[,] toInclude = s.Reserved;
            int nheight = (dir == SeamDirection.Horizontal) ? height + seams.Length : height;
            int nwidth = (dir == SeamDirection.Horizontal) ? width : width + seams.Length;
            RGBData[,] nimg = new RGBData[nheight, nwidth];
            byte[,] ngrad = new byte[nheight, nwidth];

            int bound1 = (dir == SeamDirection.Horizontal) ? width : height;
            int bound2 = (dir == SeamDirection.Horizontal) ? height : width;

            for (int i = 0; i < bound1; i++)
            {
                for (int j = 0, l = 0; j < bound2; j++, l++)
                {
                    if (dir == SeamDirection.Horizontal)
                    {
                        if (toInclude[j, i])
                        {
                            int row1 = j;
                            int row2 = (j==0) ? (j+1) : (j-1);
                            nimg[l, i].A = (byte)((img[row1, i].A + img[row2, i].A) / 2);
                            nimg[l, i].R = (byte)((img[row1, i].R + img[row2, i].R) / 2);
                            nimg[l, i].G = (byte)((img[row1, i].G + img[row2, i].G) / 2);
                            nimg[l, i].B = (byte)((img[row1, i].B + img[row2, i].B) / 2);
                            ngrad[l, i] = (byte)((grad[row1, i] + grad[row2, i]) / 2);
                            l++;
                        }
                        nimg[l, i] = img[j, i];
                        ngrad[l, i] = grad[j, i];
                    }
                    else
                    {
                        if (toInclude[i, j])
                        {
                            int col1 = j;
                            int col2 = (j == 0) ? (j + 1) : (j - 1);
                            nimg[i, l].A = (byte)((img[i, col1].A + img[i, col2].A) / 2);
                            nimg[i, l].R = (byte)((img[i, col1].R + img[i, col2].R) / 2);
                            nimg[i, l].G = (byte)((img[i, col1].G + img[i, col2].G) / 2);
                            nimg[i, l].B = (byte)((img[i, col1].B + img[i, col2].B) / 2);
                            ngrad[i, l] = (byte)((grad[i, col1] + grad[i, col2]) / 2);
                            l++;
                        }
                        nimg[i, l] = img[i, j];
                        ngrad[i, l] = grad[i, j];
                    }
                }
            }

            return new ImgPair(nimg, ngrad);
        }

        private ImgPair excludeSeamsFromImg(SeamPair s, SeamDirection dir, RGBData[,] img, byte[,] grad)
        {
            Seam[] seams = s.Seams;
            bool[,] toExclude = s.Reserved;

            int height = img.GetLength(0);
            int width = img.GetLength(1);
            int nheight = (dir == SeamDirection.Horizontal) ? height - seams.Length : height;
            int nwidth = (dir == SeamDirection.Horizontal) ? width : width - seams.Length;
            RGBData[,] nimg = new RGBData[nheight, nwidth];
            byte[,] ngrad = new byte[nheight, nwidth];

            int bound1 = (dir == SeamDirection.Horizontal) ? width : height;
            int bound2 = (dir == SeamDirection.Horizontal) ? height : width;

            for (int i = 0; i < bound1; i++)
            {
                for (int j = 0, l = 0; j < bound2; j++, l++)
                {
                    if (dir == SeamDirection.Horizontal)
                    {
                        while (j < height && toExclude[j, i])
                            j++;
                        if (j < height)
                        {
                            nimg[l, i] = img[j, i];
                            ngrad[l, i] = grad[j, i];
                        }
                    }
                    else
                    {
                        while (j < width && toExclude[i, j])
                            j++;
                        if (j < width)
                        {
                            nimg[i, l] = img[i, j];
                            ngrad[i, l] = grad[i, j];
                        }
                    }
                }
            }

            return new ImgPair(nimg, ngrad);
        }
    }
}
