using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using ImageSharp.Filtering;
using ImageSharp.Images;

namespace ImageSharp.Filtering.Linear
{
    /// <summary>
    /// Klasse fr Gradientenfilter, die zur Kantenextraktion geeignet sind.
    /// </summary>
    class GradientFilter : LinearFilter
    {
        #region Private/Protected Attribute
        protected double[,] filterMatrix2;  // 2. Filtermatrix, die anstelle der orthogonal erzeugten genutzt werden kann
        private bool getGradInfo = false;   // wenn true, dann wird zustzlich die Gradientenrichtung berechnet
        private float[,] gradOri;           // Matrix der Gradientenrichtung
        #endregion

        #region ffentliche Methoden
        public GradientFilter(double[,] matrix) : base(matrix)
        {
            createFilterMatrix2();
        }

        public GradientFilter(int[,] matrix) : base(matrix)
        {
            createFilterMatrix2();
        }

        /// <summary>
        /// Anstelle der zu matrix1 Orthogonalen wird als zweite Richtungskomponente die bergebene
        /// matrix2 verwendet.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="matrix2"></param>
        public GradientFilter(double[,] matrix1, double[,] matrix2)
            : base(matrix1)
        {
            this.filterMatrix2 = matrix2;
        }

        /// <summary>
        /// Anstelle der zu matrix1 Orthogonalen wird als zweite Richtungskomponente die bergebene
        /// matrix2 verwendet.
        /// </summary>
        /// <param name="matrix1"></param>
        /// <param name="matrix2"></param>
        public GradientFilter(int[,] matrix1, int[,] matrix2)
            : base(matrix1)
        {
            // Wandle matrix2 in double um
            this.filterMatrix2 = new double[matrix2.GetLength(0), matrix2.GetLength(1)];
            for (int i = 0; i < matrix2.GetLength(0); i++)
            {
                for (int j = 0; j < matrix2.GetLength(1); j++)
                {
                    this.filterMatrix2[i, j] = (double)matrix2[i, j];
                }
            }
        }

        /// <summary>
        /// Diese Methode liefert zum einen das Kantenbild in img zurck und gibt zum anderen die Matrix der
        /// Gradientenrichtung zurck.
        /// </summary>
        /// <param name="img">Bild, zu dem das Kantenbild berechnet werden soll</param>
        /// <returns>Matrix der Gradientenrichtung</returns>
        public float[,] GetGradInfo(ImageGray img)
        {
            bool oldVal = this.getGradInfo;
            this.getGradInfo = true;
            this.gradOri = new float[img.Height, img.Width];
            this.Apply(img);    // Gradientenfilter anwenden, dabei auch Gradientenrichtung berechnen
            this.getGradInfo = oldVal;
            return this.gradOri;
        }

        /// <summary>
        /// Wendet den Gradientenfilter auf das bergebene Bild an.
        /// </summary>
        /// <param name="img">Bild, auf den der Gradientenfilter angewendet werden soll</param>
        public override void Apply(ImageGray img)
        {
            byte[,] dataCopy = (byte[,])img.Data.Clone();

            // Wende den linearen Filter auf jedes Bildpixel an
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    double pxSumX = 0.0, pxSumY = 0.0;
                    // Gehe die Filtermatrix durch
                    for (int i = 0; i < filterMatrix.GetLength(0); i++)
                    {
                        for (int j = 0; j < filterMatrix.GetLength(1); j++)
                        {
                            int x1 = x + j - 1;
                            int y1 = y + i - 1;
                            if (x1 >= 0 && x1 < img.Width && y1 >= 0 && y1 < img.Height)
                            {
                                // summiere X- und Y-Komponenten auf
                                pxSumX += (double)dataCopy[y1, x1] * filterMatrix[i, j];
                                pxSumY += (double)dataCopy[y1, x1] * filterMatrix2[i, j];
                            }
                        }
                    }

                    // Approximiere Gradientenbetrag durch Betrge der Richtungskomponenten
                    int ival = (int)(Math.Abs(pxSumX) + Math.Abs(pxSumY));
                    if (ival > 255)
                        ival = 255;
                    img.Data[y, x] = (byte)ival;

                    // Wenn Gradientenrichtung berechnet werden soll, dann tue dies
                    if (this.getGradInfo == true)
                        this.gradOri[y, x] = (float)Math.Atan2(pxSumY, pxSumX);
                }
            }
        }
        #endregion

        #region Private Methoden
        // erzeugt die zweite Filtermatrix aus der Orthogonalen der ersten
        private void createFilterMatrix2()
        {
            if (base.filterMatrix.GetLength(0) != base.filterMatrix.GetLength(1))
                throw new Exception("Filtermatrix muss quadratisch sein...");
            this.filterMatrix2 = new double[base.filterMatrix.GetLength(0), base.filterMatrix.GetLength(1)];
            // Bilde gespiegelte Matrix aus filterMatrix1
            for (int i = 0; i < base.filterMatrix.GetLength(0); i++)
            {
                for (int j = 0; j < base.filterMatrix.GetLength(1); j++)
                {
                    this.filterMatrix2[j, i] = base.filterMatrix[i, j];
                }
            }
        }
        #endregion
    }

    /// <summary>
    /// Stellt statische get-Properties fr Gradientenfilter zur Verfgung.
    /// </summary>
    sealed class GradientFilters
    {
        #region ffentliche statische get-Properties
        /// <summary>
        /// Rckgabe eines Objekts des Sobel-Operators.
        /// </summary>
        public static GradientFilter Sobel
        {
            get
            {
                return new GradientFilter(new double[,] {
                    { -1, 0, 1 }, 
                    { -2, 0, 2 }, 
                    { -1, 0, 1} });
            }
        }

        /// <summary>
        /// Rckgabe eines Objekts des Prewitt-Operators.
        /// </summary>
        public static GradientFilter Prewitt
        {
            get
            {
                return new GradientFilter(new double[,] {
                    { -1, 0, 1 }, 
                    { -1, 0, 1 }, 
                    { -1, 0, 1 } });
            }
        }

        /// <summary>
        /// Rckgabe eines Objekts des Roberts-Operators.
        /// </summary>
        public static GradientFilter Roberts
        {
            get
            {
                return new GradientFilter(
                    new double[,] {
                        {  0, 1 }, 
                        { -1, 0 } },
                    new double[,] {
                        {  -1, 0 }, 
                        { 0, 1 } });
            }
        }
        #endregion
    }
}
