using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;
using ImageSharp.GrayValueTransforms;
using ImageSharp.Maths;

namespace ImageSharp.HistOps
{
    class HistogramOps
    {
        /// <summary>
        /// Berechnet das Histogramm des bergebenen Bildes.
        /// </summary>
        /// <param name="img">Auszuwertendes Bild</param>
        /// <returns>Histogramm-Array</returns>
        public static int[] CalcHist(ImageGray img)
        {
            int[] hist = new int[256];

            // Summiere im Bild vorkommende Grauwerte auf.
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    hist[img.Data[y, x]]++;
                }
            }

            return hist;
        }

        /// <summary>
        /// Berechnet das kumulative Histogramm des bergebenen Bildes.
        /// </summary>
        /// <param name="img">Auszuwertendes Bild</param>
        /// <returns>Histogramm-Array</returns>
        public static int[] CalcCumulativeHist(ImageGray img)
        {
            // Histogramm berechnen
            int[] hist = CalcHist(img);
            // Aufsummieren...
            for (int i = 1; i <= 255; i++)
                hist[i] += hist[i - 1];
            return hist;
        }

        /// <summary>
        /// Berechnet das kumulative Histogramm des bergebenen Histogramms.
        /// </summary>
        /// <param name="hist">Einfaches Bild-Histogramm</param>
        /// <returns>Array des kumulativen Histogramms</returns>
        public static int[] CalcCumulativeHist(int[] hist)
        {
            int[] chist = new int[hist.Length];
            // Aufsummieren...
            chist[0] = hist[0];
            for (int i = 1; i <= 255; i++)
            {
                chist[i] = chist[i - 1] + hist[i];
            }
            
            return chist;
        }
    }

    /// <summary>
    /// Fhrt einen linearen Histogrammausgleich durch.
    /// </summary>
    class HistEqualization : IGrayImgApplicable
    {
        public void Apply(ImageGray img)
        {
            // Berechne das kumulative Histogramm und die Anzahl der Pixel.
            int[] cumHist = HistogramOps.CalcCumulativeHist(img);
            long pxs = cumHist[cumHist.Length - 1];

            // Fhre den Histogrammausgleich durch.
            for (int y = 0; y < img.Height; y++)
            {
                for (int x = 0; x < img.Width; x++)
                {
                    img.Data[y, x] = (byte)(255 * cumHist[img.Data[y, x]] / pxs);
                }
            }
        }
    }

    /// <summary>
    /// Lokaler Histogrammausgleich, in derzeitiger Version sehr lahm.
    /// </summary>
    class HistLocalEqualization : IGrayImgApplicable
    {
        private int maskWidth = 3;

        public HistLocalEqualization() { }
        public HistLocalEqualization(int maskWidth)
        {
            this.maskWidth = maskWidth;
        }

        public void Apply(ImageGray img)
        {
            // Kopie der Grauwertdaten erstellen
            ImageGray img2 = new ImageGray(img);

            // Fhre fr jedes Pixel einen Histogrammausgleich in lokaler Nachbarschaft durch.
            for (int y = 0; y < img2.Height; y++)
            {
                for (int x = 0; x < img2.Width; x++)
                {
                    // Berechne lokalen Rechteckbereich.
                    int x1 = Math.Max(x - (maskWidth-1)/2, 0);
                    int y1 = Math.Max(y - (maskWidth-1)/2, 0);
                    int x2 = Math.Min(x + maskWidth/2, img2.Width - 1);
                    int y2 = Math.Min(y + maskWidth/2, img2.Height - 1);
                    int pxs = (y2 - y1 + 1) * (x2 - x1 + 1);

                    ImageGray imgLocal = new ImageGray(x2 - x1 + 1, y2 - y1 + 1);
                    for (int i = y1; i <= y2; i++)
                        for (int j = x1; j <= x2; j++)
                            imgLocal.Data[i-y1, j-x1] = img2.Data[i, j];

                    // Berechne kumulatives Histogramm.
                    int[] cumHist = HistogramOps.CalcCumulativeHist(imgLocal);

                    // Setze Pixelwert in Originalbild um.
                    img.Data[y, x] = (byte)(255 * cumHist[img2.Data[y, x]] / pxs);
                }
            }
        }
    }

    /// <summary>
    /// Fhrt eine Histogrammanpassung an das bergebene Histogramm durch.
    /// </summary>
    class HistSpecification : IGrayImgApplicable
    {
        private int[] targetHist = null;

        public HistSpecification(int[] targetHist)
        {
            this.targetHist = targetHist;
        }

        public void Apply(ImageGray img)
        {
            // Original-Histogramm berechnen
            int[] origHist = HistogramOps.CalcHist(img);
            byte[] lut = new byte[256];

            // kumulative Verteilungsfunktion von Original-Histogramm und Anpassungs-Histogramm berechnen
            double[] POrig = cdf(origHist);
            double[] PTarget = cdf(this.targetHist);

            // Lookup-Tabelle erstellen
            for (int i = 0; i <= 255; i++)
            {
                int j = 255;
                do
                {
                    lut[i] = (byte)j;
                    j--;
                } while (j >= 0 && POrig[i] <= PTarget[j]);
            }

            // LUT anwenden
            img.Apply(new LUTApplication(lut));
        }

        // Berechnet die kumulative Verteilungsfunktion des bergebenen Histogramms.
        private double[] cdf(int[] hist)
        {
            int n = 0;
            for (int i = 0; i <= 255; i++)
                n += hist[i];

            double[] P = new double[256];
            int c = 0;
            for (int i = 0; i <= 255; i++)
            {
                c += hist[i];
                P[i] = (double)c / n;
            }

            return P;
        }
    }

    /// <summary>
    /// Histogramm-Anpassung an die Gauss-Verteilung mit bergebener Standardabweichung.
    /// </summary>
    class HistSpecGauss : IGrayImgApplicable
    {
        private float sigma = 0;

        /// <summary>
        /// Konstruktor.
        /// </summary>
        /// <param name="sigma">Standardabweichung fr die Gauss-Funktion</param>
        public HistSpecGauss(float sigma)
        {
            this.sigma = sigma;
        }

        public void Apply(ImageGray img)
        {
            int[] gaussHist = new int[256];
            int pxs = img.Width * img.Height;

            // Histogramm aus Gauss-Verteilung berechnen.
            for (int i = 128; i <= 255; i++)
            {
                gaussHist[i] = gaussHist[255 - i] = (int)(MathCalc.Gauss(i - 128, sigma) * pxs);
            }

            img.Apply(new HistSpecification(gaussHist));
        }
    }
}
