using System;
using System.Collections.Generic;
using System.Text;
using ImageSharp.Images;
using System.Drawing;

namespace ImageSharp.Detections.HoughTransforms
{
    /// <summary>
    /// Speicherstruktur fr ein Element des Hough-Akkus.
    /// </summary>
    class accuElem
    {
        #region ffentliche Attribute und Eigenschaften
        /// <summary>
        /// Betrag des Akku-Elements.
        /// </summary>
        public float Amount = 0.0f;

        /// <summary>
        /// Aufsummierte Pixelanzahl.
        /// </summary>
        public int PxCount = 0;

        /// <summary>
        /// Lnge der zugehrigen Linie im Bildraum.
        /// </summary>
        public float LineLength = 0.0f;

        /// <summary>
        /// Normierter Wert, sodass lngere Linien keinen greren Einfluss haben.
        /// </summary>
        public float NormedValue
        {
            get
            {
                return (LineLength == 0f) ? 0f : Amount * PxCount / LineLength;
            }
        }
        #endregion

        #region ffentliche Methoden
        public accuElem()
        {

        }

        public accuElem(float amount, int pxCount, float lineLength)
        {
            this.Amount = amount;
            this.PxCount = pxCount;
            this.LineLength = lineLength;
        }

        /// <summary>
        /// Lscht die Daten aus dem Akku-Element.
        /// </summary>
        public void Clear()
        {
            this.Amount = 0.0f;
            this.PxCount = 0;
            this.LineLength = 0.0f;
        }

        /// <summary>
        /// Statische Methode zum Kopieren der Daten aus einem Akku-Element in ein anderes.
        /// </summary>
        /// <param name="from">Quell-Akkuelement</param>
        /// <param name="to">Ziel-Akkuelement</param>
        public static void CopyData(accuElem from, accuElem to)
        {
            to.Amount = from.Amount;
            to.PxCount = from.PxCount;
            to.LineLength = from.LineLength;
        }

        /// <summary>
        /// Kopiert die Daten aus diesem in das bergebene Akku-Element.
        /// </summary>
        /// <param name="to">Ziel-Akkuelement</param>
        public void CopyDataTo(accuElem to)
        {
            to.Amount = this.Amount;
            to.PxCount = this.PxCount;
            to.LineLength = this.LineLength;
        }

        /// <summary>
        /// Kopiert die Daten von diesem Akkuelement in die bergebene HoughLine.
        /// </summary>
        /// <param name="to">Ziel-Objekt</param>
        public void CopyDataTo(HoughLine to)
        {
            to.Accuval = this.Amount;
            to.PxCount = this.PxCount;
            to.LineLength = this.LineLength;
        }

        /// <summary>
        /// Kopiert die Daten von einem bergebenen Akkuelement in dieses.
        /// </summary>
        /// <param name="from">Quell-Akkuelement</param>
        public void CopyDataFrom(accuElem from)
        {
            this.Amount = from.Amount;
            this.PxCount = from.PxCount;
            this.LineLength = from.LineLength;
        }
        #endregion
    }

    /// <summary>
    /// Speicherstruktur fr ein Akkuelement und seine Indizes im Akku-Array.
    /// </summary>
    class maxAccuElem
    {
        #region ffentliche Attribute
        /// <summary>
        /// Winkel-Index im Akku-Array.
        /// </summary>
        public int A = 0;

        /// <summary>
        /// Radius-Index im Akku-Array.
        /// </summary>
        public int R = 0;

        /// <summary>
        /// Akku-Element.
        /// </summary>
        public accuElem Elem = null;
        #endregion

        #region ffentliche Methoden
        public maxAccuElem() { }

        public maxAccuElem(int a, int r, accuElem elem)
        {
            this.A = a;
            this.R = r;
            this.Elem = elem;
        }

        /// <summary>
        /// Lscht die Daten aus diesem Element.
        /// </summary>
        public void Clear()
        {
            this.A = 0;
            this.R = 0;
            this.Elem = null;
        }
        #endregion
    }

    /// <summary>
    /// Klasse zur Durchfhrung der Hough-Transformation fr Geraden.
    /// </summary>
    class HTLines
    {
        #region Private Attribute
        private ImageGray img = null;           // Bild, aus dem Geraden erkannt werden sollen
        private double stepAngle = 0.0;         // Winkelbereich, den ein Akku-Element umfasst
        private double stepRadius = 0.0;        // Radiusbereich, den ein Akku-Element umfasst
        private int aSteps = 0;                 // Anzahl der Winkel-Elemente im Akku (Anzahl der Zeilen)
        private int rSteps = 0;                 // Anzahl der Radius-Elemente im Akku (Anzahl der Spalten)
        private int xCenter = 0;                // x-Koordinate der Bildmitte
        private int yCenter = 0;                // y-Ordinate der Bildmitte
        private float threshPercMaxVal = 0.4f;   // Default-Schwellwert fr Akku-Elemente

        private bool useGradInfo = false;       // wenn true, dann werden die Kantenbetrge im Bild aufsummiert
        private bool useNMS = false;            // wenn true, dann wird Non-Maximum-Suppression verwendetn
        private double maxNormedAccuVal = 0.0;  // maximaler normierter Betrag im Akku-Array
        private accuElem[,] originalHoughAccu;  // Originaler aufsummierter Hough-Akku
        private accuElem[,] nmsAccu;            // Akku nach Non-Maximum-Suppression
        private accuElem[,] resultAccu;         // Resultat-Akku, z.B. nach weiterem Schwellwertverfahren
        #endregion

        #region ffentliche Eigenschaften
        /// <summary>
        /// Das zu verarbeitende Bild.
        /// </summary>
        public ImageGray Img
        {
            get
            {
                return this.img;
            }
            set
            {
                this.img = value;
            }
        }

        /// <summary>
        /// Der Schwellwert fr Akku-Elemente bei Verwendung eines Schwellwertverfahrens. Gibt an, welchen
        /// Anteil ein Akku-Element am maximalen Akku-Betrag haben muss, um als Gerade durchzugehen.
        /// </summary>
        public float ThreshPercMaxVal
        {
            get
            {
                return this.threshPercMaxVal;
            }
            set
            {
                this.threshPercMaxVal = value;
            }
        }

        /// <summary>
        /// Maximaler normierter Betrag eines Elements im Akkumulator.
        /// </summary>
        public double MaxNormedAccuVal
        {
            get
            {
                return this.maxNormedAccuVal;
            }
        }
        #endregion

        #region ffentliche Methoden
        /// <summary>
        /// Konstruktor fr die HT.
        /// </summary>
        /// <param name="aSteps">Auflsung bzgl. der Winkel (Anzahl der Zeilen des Akkus)</param>
        /// <param name="rSteps">Auflsung bzgl. des Radius (Anzahl der Spalten des Akkus)</param>
        /// <param name="img">Bild, aus dem Geraden erkannt werden sollen</param>
        /// <param name="isGradImg">Gibt an, ob das bergebene Bild ein Gradientenbild ist und diese
        /// Information genutzt werden soll</param>
        /// <param name="nms">Gibt an, ob fr die Erkennung von Linien das 
        /// Non-Maximum-Suppression-Verfahren mit anschlieender Schwellwertoperation anstatt einer
        /// ausschlielichen Schwellwertoperation durchgefhrt werden soll.</param>
        public HTLines(int aSteps, int rSteps, ImageGray img, bool isGradImg, bool nms)
        {
            // bergebene Daten speichern
            this.aSteps = aSteps;
            this.rSteps = rSteps;
            this.img = img;
            this.useGradInfo = isGradImg;
            this.useNMS = nms;

            // weitere Parameter berechnen
            this.xCenter = img.Width / 2;
            this.yCenter = img.Height / 2;
            this.stepAngle = Math.PI / aSteps;
            double rMax = Math.Sqrt(xCenter * xCenter + yCenter * yCenter);
            this.stepRadius = (2.0 * rMax) / rSteps;

            // Speicher reservieren
            nmsAccu = new accuElem[aSteps, rSteps];
            originalHoughAccu = new accuElem[aSteps, rSteps];
            resultAccu = new accuElem[aSteps, rSteps];
            for (int i = 0; i < aSteps; i++)
            {
                for (int j = 0; j < rSteps; j++)
                {
                    nmsAccu[i, j] = new accuElem();
                    originalHoughAccu[i, j] = new accuElem();
                    resultAccu[i, j] = new accuElem();
                }
            }
        }

        /// <summary>
        /// Geraden aus dem Bild erkennen.
        /// </summary>
        /// <returns>Liste erkannter Linien</returns>
        public List<Line> DetectLines()
        {
            // alle Akkus reseten
            clearAccus();

            // 1. Schritt: 2D-Akkumulator-Array fllen
            GetOriginalHoughAccu();

            // 2. Schritt: Maxima im Akku durch Schwellwertverfahren 
            //    oder Non-Maximum-Suppression finden
            if(useNMS)
                nonMaximumSuppression(this.originalHoughAccu, true);
            else
                thresholding(this.originalHoughAccu);

            // 3.) brig gebliebene Linien extrahieren
            return getLinesFromAccu(this.resultAccu);
        }

        /// <summary>
        /// Berechnet den Original-HoughAkku und gibt diesen zurck, also den ursprnglich 
        /// aufsummierten und nicht mittels Schwellwertverfahren oder 
        /// Non-Maximum-Suppression ausgewerteten.
        /// </summary>
        /// <returns>Zweidimensionales Hough-Akkumulator-Array</returns>
        public accuElem[,] GetOriginalHoughAccu()
        {
            this.maxNormedAccuVal = 0.0f;

            // gehe jedes Element des eingegrenzten Bildraums durch
            for (int i = 0; i < img.Height; i++)
            {
                for (int j = 0; j < img.Width; j++)
                {
                    // zum derzeitigen Pixel die Geraden berechnen, auf denen es liegen kann
                    // und in Hough-Akku eintragen
                    int x = j - xCenter, y = i - yCenter;
                    if (img.Data[i, j] > 0)  // Pixelwert im Kantenbild muss grer 0 sein
                    {
                        for (int a = 0; a < aSteps; a++)
                        {
                            // berechne Winkel der Geraden
                            double theta = stepAngle * (double)a;
                            // Index fr den Radius berechnen
                            int r = (int)(Math.Round((x * Math.Cos(theta) + y * Math.Sin(theta)) / stepRadius) + rSteps / 2.0);
                            if (r >= 0 && r < rSteps)
                            {
                                // wenn Lnge der Geraden noch nicht berechnet, dann tue dies
                                if (originalHoughAccu[a, r].LineLength == 0.0f)
                                {
                                    double radius = x * Math.Cos(theta) + y * Math.Sin(theta);
                                    Line line = new Line(theta, radius);
                                    //line.GetStartPoint(limits);
                                    //line.GetEndPoint(limits);
                                    // trage die Lnge in das LineLength-Attribut ein
                                    //originalHoughAccu[a, r].LineLength = (float)line.Length;
                                }
                                // inkrementiere die Anzahl der Pixel, die auf der Geraden liegen
                                originalHoughAccu[a, r].PxCount++;
                                // addiere 1 oder den Gradientenwert, wenn gewnscht
                                originalHoughAccu[a, r].Amount += (this.useGradInfo == true) ? img.Data[i, j] / 255.0f : 1;
                                
                                // speichere maximalen normierten Akku-Wert
                                if (originalHoughAccu[a, r].NormedValue > this.maxNormedAccuVal)
                                    this.maxNormedAccuVal = originalHoughAccu[a, r].NormedValue;
                            }
                        }
                    }
                }
            }

            return originalHoughAccu;
        }

        /// <summary>
        /// Gibt den Betrag des Akku-Arrays fr das Akku-Element zurck, welches mit der bergebenen
        /// Linie korrespondiert
        /// </summary>
        /// <param name="line">Gerade, deren Akku-Element-Betrag zurckgegeben werden soll</param>
        /// <returns>Akku-Betrag des entsprechenden Akku-Elements</returns>
        public double GetAccuAmountAt(Line line)
        {
            return getSomeAccuAmountAt(line, false);
        }

        /// <summary>
        /// Gibt den normierten Betrag des Akku-Elements zurck, welches der bergebenen Geraden entspricht.
        /// </summary>
        /// <param name="line">Gerade, deren normierter Akku-Betrag zurckgegeben werden soll</param>
        /// <returns>Normierter Betrag des entsprechenden Akku-Elements</returns>
        public double GetNormedAccuAmountAt(Line line)
        {
            return getSomeAccuAmountAt(line, true);
        }

        /// <summary>
        /// Gibt die Anzahl der Pixel des zu der bergebenen Geraden gehrenden Akku-Elements zurck.
        /// </summary>
        /// <param name="line">Gerade, deren Pixelanzahl ermittelt werden soll</param>
        /// <returns>Anzahl der Pixel auf der Geraden</returns>
        public int GetOriginalPixelCountAt(Line line)
        {
            // Gerade in Akku-Indizes umwandeln
            int a = (int)(line.Angle / this.stepAngle);
            int r = getRadiusIndexFromImageLine(line);

            // gib Pixelanzahl des Akku-Elements zurck
            if (a >= 0 && a < originalHoughAccu.GetLength(0) && r >= 0 && r < originalHoughAccu.GetLength(1))
                return originalHoughAccu[a, r].PxCount;
            else
                return 0;
        }

        /// <summary>
        /// Gibt den durchschnittlichen summierten Betrag der Akku-Elemente zurck, die mit der Geraden bereinstimmen oder
        /// direkt parallel dazu sind und einen Schwellwert grer threshVal * maximaler normierter 
        /// Akku-Betrag haben.
        /// </summary>
        /// <param name="line">Gerade, zu welcher und in deren Umgebung der Betrag ermittelt werden soll</param>
        /// <param name="threshVal">Schwellwert</param>
        /// <returns>Durchschnittlicher Normierter Akku-Betrag der Elemente, die der Geraden entsprechen oder direkt
        /// parallel zu dieser sind und den Schwellwert bersteigen</returns>
        public double GetNormedAmountCountAround(Line line, double threshVal)
        {
            // Akku-Indizes berechnen
            int a = (int)(line.Angle / this.stepAngle);
            int r = getRadiusIndexFromImageLine(line);

            double sumValue = 0.0;  // aufsummierter Betrag
            int win = 1;            // Fenstergre fr Parallel-Indizes
            int count = 0;          // Anzahl mit einbezogener Akku-Elemente
            if (a >= 0 && a < originalHoughAccu.GetLength(0))
            {
                for (int i = r - win; i <= r + win; i++)
                {
                    if (i >= 0 && i < originalHoughAccu.GetLength(1))
                    {
                        // wenn ber dem Schwellwert, dann summiere auf
                        if (sumValue == 0.0 || originalHoughAccu[a, i].NormedValue > threshVal * this.maxNormedAccuVal)//originalHoughAccu[a, i].Amount > maxVal)
                        {
                            sumValue += originalHoughAccu[a, i].NormedValue;
                            count++;
                        }
                    }
                }
            }

            // Wert zurckgeben
            return (count == 0) ? 0 : sumValue / count;
        }

        /// <summary>
        /// Zeichnet genau die Kurven des aktuellen Original-Akkus in das bergebene Bild ein,
        /// welche den Punkten aus Points im Akku entsprechen.
        /// </summary>
        /// <param name="points">Punkte, deren Kurven in das Bild gezeichnet werden sollen</param>
        /// <param name="img">Bild, in das der Akku-Inhalt eingezeichnet werden soll</param>
        public void DrawAccuPointsToImg(Point[] points, ImageGray img)
        {
            // setze Pixelfarbe auf wei
            for (int i = 0; i < originalHoughAccu.GetLength(0) && i < img.Width; i++)
            {
                for (int j = 0; j < originalHoughAccu.GetLength(1) && j < img.Height; j++)
                {
                    img.Data[j, i] = (byte)255;
                }
            }

            // Berechne Kurve fr jeden Punkt
            foreach (Point p in points)
            {
                // In Zentrums-Koordinaten muss gerechnet werden
                int x = p.X - xCenter;
                int y = p.Y - yCenter;
                for (int a = 0; a < aSteps && a < img.Width; a++)
                {
                    // jeweiligen Winkel und Radius berechnen sowie deren Indizes
                    double theta = stepAngle * (double)a;
                    int r = (int)(Math.Round((x * Math.Cos(theta) + y * Math.Sin(theta)) / stepRadius) + rSteps / 2.0);
                    if (r >= 0 && r < rSteps && r < img.Height)
                    {
                        // Akku-Inhalt skalieren und ins Bild einzeichnen
                        img.Data[r, a] = (byte)(byte)(255.0 - this.originalHoughAccu[a, r].NormedValue * 255.0 / maxNormedAccuVal);
                    }
                }
            }
        }

        /// <summary>
        /// Inhalt des aktuellen Original-Akkus wird in das bergebene Bild eingezeichnet.
        /// </summary>
        /// <param name="img">Bild, in das der Akku-Inhalt gezeichnet werden soll</param>
        public void DrawAccuToImg(ImageGray img)
        {
            for (int i = 0; i < originalHoughAccu.GetLength(0) && i < img.Width; i++)
            {
                for (int j = 0; j < originalHoughAccu.GetLength(1) && j < img.Height; j++)
                {
                    img.Data[j, i] = (byte)(255.0 - this.resultAccu[i, j].NormedValue * 255.0 / maxNormedAccuVal);
                }
            }
        }
        #endregion

        #region Private Methoden
        // Fhrt auf dem bergebenen Hough-Akku das Non-Maximum-Suppression Verfahren durch,
        // um lokale Maxima zu ermitteln.
        // param "accu": Akku, auf dem das NMS-Verfahren durchgefhrt werden soll
        // param "withThresholding": Gibt an, ob anschlieend noch ein Schwellwertverfahren
        // durchzufhren ist
        // returns: Neues Akkumulator-Array, dieses ist zudem in dem Attribut nmsAccu gespeichert
        private accuElem[,] nonMaximumSuppression(accuElem[,] accu, bool withThresholding)
        {
            // gehe jedes Element des Akkus durch
            for (int i = 0; i < aSteps; i++)
            {
                // berechne Indizes des oberen und unteren Nachbarelements, soweit vorhanden
                int klow = (i == 0) ? 0 : -1;
                int khigh = (i == aSteps - 1) ? 0 : 1;
                for (int j = 0; j < rSteps; j++)
                {
                    accu[i, j].CopyDataTo(nmsAccu[i, j]);
                    bool next = false;
                    // berechne Indizes des linken und rechten Nachbarelements, soweit vorhanden
                    int llow = (j == 0) ? 0 : -1;
                    int lhigh = (j == rSteps - 1) ? 0 : 1;

                    // Schwellwertverfahren, wenn gewnscht
                    // wenn Wert zu klein, dann schliee das Pixel aus
                    if (withThresholding && nmsAccu[i, j].NormedValue < this.threshPercMaxVal * this.maxNormedAccuVal)
                    {
                        nmsAccu[i, j].Clear();
                    }
                    else
                    {
                        // eigentliches Non-Maximum-Suppression
                        for (int k = klow; k <= khigh && !next; k++)
                        {
                            for (int l = llow; l <= lhigh && !next; l++)
                            {
                                // wenn ein Nachbarelement einen greren Wert hat, so
                                // schliee aktuelles Element aus neuem Akku aus
                                if (accu[i + k, j + l].Amount > accu[i, j].Amount)
                                {
                                    nmsAccu[i, j].Clear();
                                    next = true;
                                }
                            }
                        }
                    }
                }
            }

            // Resultat-Akku setzen und zurckgeben
            resultAccu = nmsAccu;
            return resultAccu;
        }

        // Fhrt ein einfaches Schwellwertverfahren auf dem bergebenen Akku aus. Setzt 
        // resultAccu auf den verarbeiteten Akku und gibt diesen zurck.
        private accuElem[,] thresholding(accuElem[,] accu)
        {
            for (int i = 0; i < aSteps; i++)
            {
                for (int j = 0; j < rSteps; j++)
                {
                    // Wenn Wert ber Schwellwert, dann behalte ihn
                    if (accu[i, j].NormedValue >= this.threshPercMaxVal * this.maxNormedAccuVal)
                    {
                        accu[i, j].CopyDataTo(resultAccu[i, j]);
                    }
                }
            }

            return resultAccu;
        }

        // Gibt den Akku-Betrag bzw. wenn gewnscht den normalisierten Akku-Betrag zu der bergebenen
        // Linie zurck.
        private double getSomeAccuAmountAt(Line line, bool normed)
        {
            // Akku-Indizes aus der Geraden berechnen
            int a = (int)(line.Angle / this.stepAngle);
            int r = getRadiusIndexFromImageLine(line);

            // (normierten) Betrag des entsprechenden Akku-Elements zurckgeben
            if (a >= 0 && a < originalHoughAccu.GetLength(0) && r >= 0 && r < originalHoughAccu.GetLength(1))
                return normed ? originalHoughAccu[a, r].NormedValue : originalHoughAccu[a, r].Amount;
            else
                return 0.0;
        }

        // Gibt den Index des Radius (Spalte im Hough-Akku) zurck, der mit der bergebenen
        // Gerade bereinstimmt.
        private int getRadiusIndexFromImageLine(Line line)
        {
            PointF p1 = new PointF();
            
            // 1.) 1 Punkt auf der Geraden in Bildkoordinaten berechnen
            p1.Y = 0.0f;
            p1.X = (float)((line.Radius - p1.Y * Math.Sin(line.Angle)) / Math.Cos(line.Angle));

            // 2.) ermittelten Punkt in Centerkoordinaten umrechnen
            p1.X = p1.X - this.xCenter;
            p1.Y = p1.Y - this.yCenter;

            // 3.) Radius + dessen Index bestimmen
            double radius = p1.X * Math.Cos(line.Angle) + p1.Y * Math.Sin(line.Angle);
            return (int)(radius / this.stepRadius + this.rSteps / 2.0);
        }

        // Erzeugt die Gerade, welche mit dem Akku-Element korrespondiert, dessen Indizes
        // bergeben wurden und gibt diese zurck.
        private Line getImageLineFromAccuElem(int a, int r)
        {
            double theta = stepAngle * (double)a;
            double radius = ((double)r - rSteps / 2.0) * stepRadius;
            PointF p1 = new PointF();

            // 1.) 1 Punkt auf der Geraden in Center-Koordinaten berechnen
            p1.Y = 0.0f;
            p1.X = (float)((radius - p1.Y * Math.Sin(theta)) / Math.Cos(theta));

            // 2.) ermittelten Punkt in Bildkoordinaten umrechnen
            p1.X = p1.X + this.xCenter;
            p1.Y = p1.Y + this.yCenter;

            // 3.) Radius in Bildkoordinaten bestimmen
            radius = p1.X * Math.Cos(theta) + p1.Y * Math.Sin(theta);

            return new Line(theta, radius);
        }

        // Erzeugt zu allen im bergebenen Akku enthaltenen Elementen != 0 die entsprechenden Geraden
        // und gibt eine Liste dieser zurck.
        private List<Line> getLinesFromAccu(accuElem[,] accu)
        {
            // Zurckzugebende Geraden
            List<Line> lines = new List<Line>();
            for (int i = 0; i < aSteps; i++)
            {
                for (int j = 0; j < rSteps; j++)
                {
                    // Wenn Element im Akku eingetragen, dann fge Linie hinzu
                    if (accu[i, j].Amount > 0)
                    {
                        lines.Add(getImageLineFromAccuElem(i, j));
                    }
                }
            }

            // Gib die Geraden zurck
            return lines;
        }

        // Setzt alle Akkus der HT zurck.
        private void clearAccus()
        {
            // Zurcksetzen jedes Akku-Elements
            for (int i = 0; i < aSteps; i++)
            {
                for (int j = 0; j < rSteps; j++)
                {
                    nmsAccu[i, j].Clear();
                    originalHoughAccu[i, j].Clear();
                    resultAccu[i, j].Clear();
                }
            }
            this.maxNormedAccuVal = 0.0;
        }
        #endregion
    }

    class HTCircles
    {
        
    }
}
