using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;
using ImageSharp.Maths;
using ImageSharp.Filtering;

namespace ImageSharp.Images
{
    /// <summary>
    /// Bild, dessen Daten komplexe Zahlen darstellen. Ntig zur Durchfhrung der Fourier-Transformation.
    /// </summary>
    class ImageComplex : IImage
    {
        #region Private Attribute
        private int height = 0, width = 0;          // Hhe und Breite des Bildes
        private int heightOrg = 0, widthOrg = 0;    // Original-Bildhhe und -Breite
        private bool fmode = false;                 // Gibt an, ob das Bild gerade fourier-transformiert ist
        private Complex[,] data = null;             // Pixeldaten
        #endregion

        #region ffentliche Eigenschaften
        /// <summary>
        /// Rckgabe der Bildhhe, die eine Zweierpotenz darstellt.
        /// </summary>
        public int Height
        {
            get
            {
                return this.height;
            }
        }

        /// <summary>
        /// Rckgabe der Bildbreite, die eine Zweierpotenz darstellt.
        /// </summary>
        public int Width
        {
            get
            {
                return this.width;
            }
        }

        /// <summary>
        /// Rckgabe der Original-Bildhhe.
        /// </summary>
        public int HeightOrg
        {
            get
            {
                return this.heightOrg;
            }
        }

        /// <summary>
        /// Rckgabe der Original-Bildbreite.
        /// </summary>
        public int WidthOrg
        {
            get
            {
                return this.widthOrg;
            }
        }

        /// <summary>
        /// true, wenn sich das Bild gerade im Fourier-Raum befindet.
        /// </summary>
        public bool Fmode
        {
            get
            {
                return this.fmode;
            }
        }

        /// <summary>
        /// Gibt die Daten des Bildes zurck.
        /// </summary>
        public Complex[,] Data
        {
            get
            {
                return this.data;
            }
        }
        #endregion

        #region ffentliche Methoden
        /// <summary>
        /// Erzeugung eines komplexen Bildes aus dem bergebenen Bitmap heraus.
        /// </summary>
        /// <param name="img">Bild, aus dem das Komplexe zu erstellen ist</param>
        public unsafe ImageComplex(Bitmap img)
        {
            // Original-Bildgre speichern
            this.heightOrg = img.Height;
            this.widthOrg = img.Width;

            // Bildgre muss 2er-Potenz sein, daher ist diese zu berechnen
            int i = 1;
            while (i < this.heightOrg)
            {
                i *= 2;
            }
            this.height = i;

            i = 1;
            while (i < this.widthOrg)
            {
                i *= 2;
            }
            this.width = i;

            // Datenfeld anlegen
            this.data = new Complex[this.height, this.width];

            // Bilddaten kopieren
            GraphicsUnit gUnit = GraphicsUnit.Pixel;
            RectangleF fBounds = img.GetBounds(ref gUnit);
            Rectangle bounds = new Rectangle((int)fBounds.X, (int)fBounds.Y,
                (int)fBounds.Width, (int)fBounds.Height);

            BitmapData bitData = img.LockBits(
                bounds, ImageLockMode.ReadOnly, PixelFormat.Format32bppRgb);
            RGBData* pBase = (RGBData*)bitData.Scan0.ToPointer();

            for (int y = 0; y < this.heightOrg; y++)
            {
                for (int x = 0; x < this.widthOrg; x++)
                {
                    RGBData* pPixel = pBase + x + y * bounds.Width;
                    // Pixel als Grauwerte speichern -> auf 1 normieren und in Realteil eintragen
                    data[y, x].Re = (0.6375f * pPixel->R + 2.1462f * pPixel->G + 0.2163f * pPixel->B) / (3 * 255f);
                }
            }

            img.UnlockBits(bitData);
        }

        /// <summary>
        /// Erzeugt ein komplexes Bild aus dem bergebenen Grauwertbild.
        /// </summary>
        /// <param name="img">Zu kopierendes Bild</param>
        public ImageComplex(ImageGray img)
        {
            this.heightOrg = img.Height;
            this.widthOrg = img.Width;

            // Hhe und Breite als Zweierpotenz erzeugen
            int i = 1;
            while (i < this.heightOrg)
            {
                i *= 2;
            }
            this.height = i;

            i = 1;
            while (i < this.widthOrg)
            {
                i *= 2;
            }
            this.width = i;

            // Datenfeld anlegen
            this.data = new Complex[this.height, this.width];
            
            // Daten kopieren
            for (int y = 0; y < this.heightOrg; y++)
            {
                for (int x = 0; x < this.widthOrg; x++)
                {
                    // Wert = Realteil auf 1 normiert
                    this.data[y, x].Re = img.Data[y, x] / 255f;
                }
            }
        }

        /// <summary>
        /// Erzeugt ein komplexes Bild aus dem bergebenen Bild als tiefe Kopie.
        /// </summary>
        /// <param name="img">Zu kopierendes Bild</param>
        public ImageComplex(ImageComplex img)
        {
            this.height = img.Height;
            this.width = img.Width;
            this.heightOrg = img.HeightOrg;
            this.widthOrg = img.WidthOrg;
            this.fmode = img.Fmode;
            this.data = (Complex[,])img.Data.Clone();
        }

        /// <summary>
        /// Konvertiert die aktuellen Bilddaten in ein .NET Bitmap-Objekt.
        /// </summary>
        /// <returns>Konvertiertes Bitmap</returns>
        public unsafe Bitmap ToBitmap()
        {
            // wenn in Ortsraum, dann ist die Bildgre = originaler Bildgre, ansonsten Gre
            // des Ortsfrequenzraumes
            int h = heightOrg, w = widthOrg;
            if (fmode)
            {
                h = height;
                w = width;
            }
            
            // neues Bitmap erzeugen
            Bitmap dstImg = new Bitmap(w, h);

            // Daten kopieren
            GraphicsUnit gUnit = GraphicsUnit.Pixel;
            RectangleF fBounds = dstImg.GetBounds(ref gUnit);
            Rectangle bounds = new Rectangle((int)fBounds.X, (int)fBounds.Y,
                (int)fBounds.Width, (int)fBounds.Height);

            BitmapData bitData = dstImg.LockBits(
                bounds, ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb);
            RGBData* pBase = (RGBData*)bitData.Scan0.ToPointer();

            double scale = (this.fmode) ? (double)Math.Sqrt(w * h) : 1;

            for (int y = 0; y < h; y++)
            {
                for (int x = 0; x < w; x++)
                {
                    RGBData* pPixel = pBase + x + y * bounds.Width;
                    // Pixelwert entspricht dem skalierten Betrag der jeweiligen komplexen Zahl
                    byte val = (byte)Math.Max(0, Math.Min(255, this.data[y, x].Magnitude * scale * 255));
                    pPixel->R = val;
                    pPixel->G = val;
                    pPixel->B = val;
                }
            }

            dstImg.UnlockBits(bitData);
            return dstImg;
        }

        public IImage Clone()
        {
            return new ImageComplex(this);
        }

        /// <summary>
        /// Vorwrtsgerichtete Fourier-Transformation durchfhren, um das Bild vom Ortsraum in den
        /// Ortsfrequenzraum zu transformieren.
        /// </summary>
        public void MakeForwardFourierTransform()
        {
            if (!fmode)
            {
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        if (((x + y) & 0x1) != 0)
                        {
                            data[y, x].Re *= -1;
                            data[y, x].Im *= -1;
                        }
                    }
                }

                // Zweidimensionale FFT durchfhren
                MathCalc.FFT2(data, FourierDirection.Forward);
                fmode = true;
            }
        }

        /// <summary>
        /// Rckwrtsgerichtete Fourier-Transformation durchfhren, um das Bild vom Ortsfrequenzraum
        /// zurck in den Ortsraum zu berfhren.
        /// </summary>
        public void MakeBackwardFourierTransform()
        {
            if (fmode)
            {
                // Zweidimensionale FFT durchfhren
                MathCalc.FFT2(data, FourierDirection.Backward);
                fmode = false;

                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        if (((x + y) & 0x1) != 0)
                        {
                            data[y, x].Re *= -1;
                            data[y, x].Im *= -1;
                        }
                    }
                }
            }
        }

        public void Apply(IComplexImgApplicable filter)
        {
            filter.Apply(this);
        }
        #endregion

        #region interface implementations
        public IPixelData GetValueAt(int x, int y)
        {
            return new ComplexData(this.data[y, x]);
        }

        public void SetValueAt(int x, int y, IPixelData value)
        {
            this.data[y, x] = ((ComplexData)value).Value;
        }

        public void CopyFrom(IImage img)
        {
            ImageComplex nimg = null;
            if (img is ImageComplex)
                nimg = new ImageComplex((ImageComplex)img);
            else if (img is ImageGray)
            {
                nimg = new ImageComplex((ImageGray)img);
            }
            else if (img is ImageColor)
                nimg = new ImageComplex(new ImageGray((ImageColor)img));

            this.data = (Complex[,])nimg.data.Clone();
            this.height = nimg.height;
            this.width = nimg.width;
        }
        #endregion
    }
}
