using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Drawing.Imaging;

namespace ImageSharp.Images
{
    /// <summary>
    /// Klasse fr ein Grauwertbild. Auf diesem knnen anschlieend alle diversen implementierten
    /// Bildverarbeitungs-Operationen ausgefhrt werden.
    /// </summary>
    class ImageGray : IImage
    {
        #region Private Attribute
        private int height = 0, width = 0;  // Hhe und Breite des Bildes
        private byte[,] data = null;        // Pixelwerte
        #endregion

        #region ffentliche Eigenschaften
        /// <summary>
        /// Gibt die Bildhhe zurck.
        /// </summary>
        public int Height
        {
            get
            {
                return this.height;
            }
        }

        /// <summary>
        /// Gibt die Bildbreite zurck.
        /// </summary>
        public int Width
        {
            get
            {
                return this.width;
            }
        }

        /// <summary>
        /// Gibt die Grauwert-Daten der einzelnen Pixel zurck.
        /// </summary>
        public byte[,] Data
        {
            get
            {
                return this.data;
            }
            set
            {
                if (value.Length != this.data.Length)
                {
                    throw new FormatException("Data lengths must be the same");
                }
                for (int i = 0; i < this.data.GetLength(0); i++)
                    for (int j = 0; j < this.data.GetLength(1); j++)
                        this.data[i, j] = value[i, j];
            }
        }
        #endregion

        #region ffentliche Methoden
        #region Konstruktoren
        /// <summary>
        /// Erzeugt ein Grauwertbild mit den bergebenen Grauwert-Daten. Dieses wird nicht tief kopiert,
        /// sondern nur die Referenz darauf bernommen.
        /// </summary>
        /// <param name="data">Zu bernehmendes Daten-Array</param>
        public ImageGray(byte[,] data)
        {
            this.height = data.GetLength(0);
            this.width = data.GetLength(1);
            this.data = data;
        }

        /// <summary>
        /// Erzeugt ein Grauwertbild mit den bergebenen Grauwert-Daten, die
        /// als Gleitkomma-Array bergeben werden. Diese werden als tiefe Kopie erstellt,
        /// das bergebene Array wird also nachher nicht mehr bentigt.
        /// </summary>
        /// <param name="data">Zu bernehmende Daten</param>
        public ImageGray(float[,] data)
        {
            this.height = data.GetLength(0);
            this.width = data.GetLength(1);
            this.data = new byte[this.height, this.width];
            for (int i = 0; i < this.height; i++)
            {
                for (int j = 0; j < this.width; j++)
                {
                    this.data[i, j] = (byte)Math.Max(0, Math.Min(255, data[i, j]));
                }
            }
        }

        /// <summary>
        /// Erzeugt ein "leeres" Grauwertbild in der bergebenen Gre.
        /// </summary>
        /// <param name="width">Bildbreite</param>
        /// <param name="height">Bildhhe</param>
        public ImageGray(int width, int height)
        {
            this.height = height;
            this.width = width;
            this.data = new byte[this.height, this.width];
        }

        /// <summary>
        /// Erzeugt ein neues Grauwertbild als tiefe Kopie des bergebenen Grauwert-Bildes.
        /// </summary>
        /// <param name="img">Zu klonendes Grauwertbild</param>
        public ImageGray(ImageGray img)
        {
            this.height = img.Height;
            this.width = img.Width;
            this.data = (byte[,])img.Data.Clone();
        }

        /// <summary>
        /// Erzeugt ein Grauwertbild aus dem bergebenen .NET Bitmap-Objekt.
        /// </summary>
        /// <param name="img">Bild, dessen Grauwertbild erzeugt werden soll</param>
        public unsafe ImageGray(Bitmap img)
        {
            // Bildgre bernehmen und Datenfeld allokieren
            this.height = img.Height;
            this.width = img.Width;
            this.data = new byte[this.height, this.width];

            // Pixeldaten kopieren
            GraphicsUnit gUnit = GraphicsUnit.Pixel;
            RectangleF fBounds = img.GetBounds(ref gUnit);
            Rectangle bounds = new Rectangle((int)fBounds.X, (int)fBounds.Y,
                (int)fBounds.Width, (int)fBounds.Height);

            BitmapData bitData = img.LockBits(
                bounds, ImageLockMode.ReadOnly, PixelFormat.Format32bppRgb);
            RGBData* pBase = (RGBData*)bitData.Scan0.ToPointer();

            for (int y = 0; y < bounds.Height; y++)
            {
                for (int x = 0; x < bounds.Width; x++)
                {
                    RGBData* pPixel = pBase + x + y * bounds.Width;
                    // RGB in Grauwerte umrechnen und als byte speichern
                    data[y, x] = (byte)((byte)((0.6375f * pPixel->R + 2.1462f * pPixel->G + 0.2163f * pPixel->B) / 3.0));
                }
            }

            img.UnlockBits(bitData);
        }

        /// <summary>
        /// Erzeugt ein Grauwertbild aus dem bergebenen ImageColor-Objekt.
        /// </summary>
        /// <param name="img">Bild, dessen Grauwertbild erzeugt werden soll</param>
        public unsafe ImageGray(ImageColor img)
        {
            // Bildgre bernehmen und Datenfeld allokieren
            this.height = img.Height;
            this.width = img.Width;
            this.data = new byte[this.height, this.width];

            // Pixeldaten kopieren
            for (int y = 0; y < this.height; y++)
            {
                for (int x = 0; x < this.width; x++)
                {
                    this.data[y, x] = (byte)((byte)((0.6375f * img.Data[y, x].R + 2.1462f * img.Data[y, x].G + 0.2163f * img.Data[y, x].B) / 3.0));
                }
            }
        }

        /// <summary>
        /// Erzeugt ein Grauwertbild aus dem bergebenen komplexen Bild.
        /// </summary>
        /// <param name="img">Komplexes Bild, dessen Grauwertbild erstellt werden soll</param>
        public ImageGray(ImageComplex img)
        {
            // Original-Gre bernehmen wenn in Ortsraum, Fourierraum-Gre sonst
            int h = img.HeightOrg, w = img.WidthOrg;
            if (img.Fmode)
            {
                h = img.Height;
                w = img.Width;
            }

            // Gre speichern und Datenfeld allokieren
            this.height = h;
            this.width = w;
            this.data = new byte[this.height, this.width];

            // Skalierungsfaktor festlegen
            float scale = (img.Fmode) ? (float)Math.Sqrt(this.width * this.height) : 1;

            for (int y = 0; y < this.height; y++)
            {
                for (int x = 0; x < this.width; x++)
                {
                    // Pixel-Wert = Skalierter Betrag des Betrags der jeweiligen komplexen Zahl
                    this.data[y, x] = (byte)Math.Max(0, Math.Min(255, img.Data[y, x].Magnitude * scale * 255f));
                }
            }
        }
        #endregion

        #region Umwandlungen eines ImageGray-Objekts
        /// <summary>
        /// Erzeugt ein .NET Bitmap Objekt aus den hier enthaltenen Grauwert-Daten.
        /// </summary>
        /// <returns>Bitmap-Objekt</returns>
        public unsafe Bitmap ToBitmap()
        {
            // Erzeuge das Bitmap
            Bitmap dstImg = new Bitmap(this.width, this.height);

            // Kopiere die Daten
            GraphicsUnit gUnit = GraphicsUnit.Pixel;
            RectangleF fBounds = dstImg.GetBounds(ref gUnit);
            Rectangle bounds = new Rectangle((int)fBounds.X, (int)fBounds.Y,
                (int)fBounds.Width, (int)fBounds.Height);

            BitmapData bitData = dstImg.LockBits(
                bounds, ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb);
            RGBData* pBase = (RGBData*)bitData.Scan0.ToPointer();

            for (int y = 0; y < bounds.Height; y++)
            {
                for (int x = 0; x < bounds.Width; x++)
                {
                    RGBData* pPixel = pBase + x + y * bounds.Width;
                    pPixel->R = this.data[y, x];
                    pPixel->G = this.data[y, x];
                    pPixel->B = this.data[y, x];
                }
            }

            dstImg.UnlockBits(bitData);
            return dstImg;
        }

        public IImage Clone()
        {
            return new ImageGray(this);
        }
        #endregion

        #region Filteranwendung
        public void Apply(IGrayImgApplicable op)
        {
            op.Apply(this);
        }
        #endregion

        #region interface implementations
        public IPixelData GetValueAt(int x, int y)
        {
            return new GrayData(this.data[y, x]);
        }

        public void SetValueAt(int x, int y, IPixelData value)
        {
            this.data[y, x] = ((GrayData)value).Value;
        }

        public void CopyFrom(ImageGray img)
        {
            this.data = new byte[img.height, img.width];
            for (int y = 0; y < img.height; y++)
                for (int x = 0; x < img.width; x++)
                    this.data[y, x] = img.data[y, x];
            this.height = img.height;
            this.width = img.width;
        }

        public void CopyFrom(byte[,] data)
        {
            int height = data.GetLength(0);
            int width = data.GetLength(1);
            this.data = new byte[height, width];
            for (int y = 0; y < height; y++)
                for (int x = 0; x < width; x++)
                    this.data[y, x] = data[y, x];
            this.height = height;
            this.width = width;
        }

        public void CopyFrom(IImage img)
        {
            ImageGray nimg = null;
            if (img is ImageGray)
                nimg = new ImageGray((ImageGray)img);
            else if (img is ImageColor)
                nimg = new ImageGray((ImageColor)img);
            else if (img is ImageComplex)
                nimg = new ImageGray((ImageComplex)img);

            this.data = (byte[,])nimg.data.Clone();
            this.height = nimg.height;
            this.width = nimg.width;
        }

        public void CreateFrom(int width, int height)
        {
            this.width = width;
            this.height = height;
            this.data = new byte[height, width];
        }
        #endregion
        #endregion
    }
}
