using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace ImageSharp.Maths
{
    /// <summary>
    /// Enumeration fr die Richtung der Fourier-Transformation.
    /// </summary>
    public enum FourierDirection
    {
        Forward = 1,
        Backward = -1
    }
    
    /// <summary>
    /// Stellt mathematische Operationen zur Verfgung, die als statische Methoden genutzt werden knnen.
    /// </summary>
    class MathCalc
    {
        #region ffentliche Methoden
        /// <summary>
        /// Schnittpunkt der mittels Winkel und Radius in HNF angegebenen Linie mit der Ordinate y berechnen.
        /// </summary>
        /// <param name="angle">Winekl der Geraden</param>
        /// <param name="radius">Radius der Geraden</param>
        /// <param name="y">Ordinate, mit der der Schnittpunkt berechnet werden soll</param>
        /// <returns>Schnittpunkt-Koordinate x der Geraden mit y</returns>
        static public int GetInterceptCoord(double angle, double radius, int y)
        {
            return (int)Math.Round((radius - (double)y * Math.Sin(angle)) / Math.Cos(angle));
        }

        /// <summary>
        /// Schnittpunkt der mittels zweier Punkte bergebenen Linie mit der Ordinate y berechnen.
        /// </summary>
        /// <param name="p1">Erster Geradenpunkt</param>
        /// <param name="p2">Zweiter Geradenpunkt</param>
        /// <param name="y">Schnittpunkt-Ordinate</param>
        /// <returns>Koordinate x des Schnittpunktes der Geraden mit y</returns>
        static public int GetInterceptCoord(Point p1, Point p2, int y)
        {
            // Berechne HNF der Geraden und rufe dann die HNF-Version auf.
            double angle = Math.PI/2.0 + Math.Atan2(p2.Y - p1.Y, p2.X - p1.X);
            double radius = p1.X * Math.Cos(angle) + p1.Y * Math.Sin(angle);
            return GetInterceptCoord(angle, radius, y);
        }

        /// <summary>
        /// Wandelt die als String bergebene Zahl in double um. Dabei kann als Dezimaltrennzeichen
        /// ein Komma oder ein Punkt dienen. Parse.Double funktioniert hier nicht, daher die
        /// Eigenentwicklung.
        /// </summary>
        /// <param name="number">String, der die zu parsende Zahl darstellt. Darf nur Zahlen, Punkt oder Komma
        /// enthalten.</param>
        /// <returns>Geparste double-Zahl oder Double.NaN bei Misserfolg.</returns>
        static public double ParseDouble(string number)
        {
            number = number.Trim();
            double dno = 0;

            if (number != "")
            {
                // Berechne erstes Vorkommen eines Punktes oder Kommas.
                int iPoint = number.IndexOf('.');
                int iComma = number.IndexOf(',');
                int firstSnip = (iPoint != -1 && iComma != -1) ? Math.Min(iPoint, iComma) : Math.Max(iPoint, iComma);
                // Berechne letztes Vorkommen eines Punktes oder Kommas.
                iPoint = number.LastIndexOf('.');
                iComma = number.LastIndexOf(',');
                int lastSnip = Math.Max(iPoint, iComma);
                string intPlace = number;
                string decPlace = "0";
                // Wandle alles vor erstem Komma/Punkt in Vorkommastellen und alles nach letztem Komma/Punkt
                // in Nachkommastellen um.
                if (firstSnip != -1)
                    intPlace = number.Substring(0, firstSnip);
                if (lastSnip != -1)
                    decPlace = number.Substring(lastSnip + 1);
                if (intPlace == "")
                    intPlace = "0";
                if (decPlace == "")
                    decPlace = "0";
                try
                {
                    dno = Double.Parse(intPlace) + Double.Parse(decPlace) / Math.Pow(10, decPlace.Length);
                }
                catch
                {
                    return Double.NaN;
                }
            }

            return dno;
        }
        
        
        /// <summary>
        /// Berechnet die mittlere Abweichung der Gauss-Fkt. (in Abhngigkeit von sigma) von 0 an der Stelle x.
        /// </summary>
        /// <param name="x">Funktionsparameter</param>
        /// <param name="sigma">Standardabweichung</param>
        /// <returns>Mittelwert der Gauss-Funktion</returns>
        static public float MeanGauss(float x, float sigma)
        {
            return (Gauss(x, sigma) + Gauss(x + 0.5f, sigma) + Gauss(x - 0.5f, sigma))
                / (3.0f * (float)Math.PI * 2.0f * sigma * sigma);
        }

        /// <summary>
        /// Berechnet den Wert der Gauss-Fkt. (in Abh. von sigma) an der Stelle x.
        /// </summary>
        /// <param name="x">Funktionsparameter</param>
        /// <param name="sigma">Standardabweichung</param>
        /// <returns>Wert der Gauss-Funktion</returns>
        static public float Gauss(float x, float sigma)
        {
            return (float)Math.Exp((double)((-x * x) / (2.0f * sigma * sigma)));
        }

        /// <summary>
        /// Erzeugt die Gauss-Maske mit der angegebenen Standardabweichung.
        /// </summary>
        /// <param name="sigma">Standardabweichung der Gauss-Funktion</param>
        /// <returns>2D-Gauss-Maske</returns>
        static public double[,] GaussMask(double sigma)
        {
            // Maske soll max. 11 Pixel breit werden
            float[] gauMax = new float[6];
            int hwidth = 6;
            int width = 11;

            // Erzeuge Gauss-Filtermaske mit den Werten der Gauss-Funktion.
            for (int i = 0; i < 6; i++)
            {
                gauMax[i] = MathCalc.Gauss(i, (float)sigma);
                if (gauMax[i] < 0.005f) // brich ab, wenn Fehler gering genug wird
                {
                    hwidth = i + 1;
                    width = i * 2 + 1;
                    break;
                }
            }

            // Dopplung der einen Maskenhlfte auf ganze Maskenbreite.
            float[] gau = new float[width];
            gau[hwidth - 1] = gauMax[0];
            for (int i = 0; i < hwidth - 1; i++)
                gau[i] = gau[width - i - 1] = gauMax[hwidth - i - 1];

            // Erzeugung der 2D-Resultatmaske durch Multiplikation der 1D-Maske in beide Richtungen.
            double[,] resultMask = new double[width, width];
            for (int i = 0; i < width; i++)
                for (int j = 0; j < width; j++)
                    resultMask[i, j] = gau[i] * gau[j];

            return resultMask;
        }
        
        /// <summary>
        /// Eigene Vorzeichenfunktion, die 1 fr Werte >= 0 und -1 fr andere Werte zurckgibt.
        /// </summary>
        /// <param name="val">Funktionsparameter</param>
        /// <returns>1 fr val>=0, -1 sonst</returns>
        static public int SignOwn(int val)
        {
            return (val >= 0) ? 1 : -1;
        }
        
        /// <summary>
        /// Lst das durch mat und b bergebene lineare Gleichungssystem mittels QR-Zerlegung,
        /// liefert entweder die exakte oder eine Ausgleichslsung.
        /// </summary>
        /// <param name="mat">Linke Teilmatrix des LGS</param>
        /// <param name="b">Rechter Vektor des LGS</param>
        /// <returns>Lsungsvektor des mittels QR-Zerlegung gelsten LGS</returns>
        static public double[] SolveLinEqnSystem(double[,] mat, double[] b)
        {
            int m = mat.GetLength(0);
            int n = mat.GetLength(1);
            double[,] A = new double[m, n+1];
            double[] R = new double[n];
            double[] L = new double[n];
            
            // A = mat|b erzeugen.
            for (int i = 0; i < m; i++)
            {
                for (int j = 0; j <= n; j++)
                {
                    A[i, j] = ((j == n) ? b[i] : mat[i, j]);
                }
            }

            // QR-Zerlegung mittels Householder-Transformation berechnen.
            house(A, R);

            // Lsungsvektor durch Rcksubstitution ermitteln.
            for (int i = n - 1; i >= 0; i--)
            {
                L[i] = A[i, n];
                for (int j = i + 1; j < n; j++)
                    L[i] -= A[i, j] * L[j];
                L[i] /= R[i];
            }

            return L;
        }

        /// <summary>
        /// Fhrt eine 1D-Fourier-Transformation durch.
        /// </summary>
        /// <param name="data">Zu transformierende Daten</param>
        /// <param name="direction">Richtung der Transformation</param>
        public static void DFT(Complex[] data, FourierDirection direction)
        {
            int n = data.Length;
            double arg, cos, sin;
            Complex[] dst = new Complex[n];

            // fr jedes Zielelement...
            for (int i = 0; i < n; i++)
            {
                dst[i] = Complex.Zero;

                arg = -(int)direction * 2.0 * System.Math.PI * (double)i / (double)n;

                // Summiere Quell-Elemente
                for (int j = 0; j < n; j++)
                {
                    cos = System.Math.Cos(j * arg);
                    sin = System.Math.Sin(j * arg);

                    dst[i].Re += (float)(data[j].Re * cos - data[j].Im * sin);
                    dst[i].Im += (float)(data[j].Re * sin + data[j].Im * cos);
                }
            }

            // Kopiere die Zielelemente
            if (direction == FourierDirection.Forward)
            {
                for (int i = 0; i < n; i++)
                {
                    data[i].Re = dst[i].Re / n;
                    data[i].Im = dst[i].Im / n;
                }
            }
            else
            {
                for (int i = 0; i < n; i++)
                {
                    data[i].Re = dst[i].Re;
                    data[i].Im = dst[i].Im;
                }
            }
        }

        /// <summary>
        /// 2D-Diskrete Fourier Transformation.
        /// </summary>
        /// <param name="data">Zu transformierende Daten</param>
        /// <param name="direction">Transformationsrichtung</param>
        public static void DFT2(Complex[,] data, FourierDirection direction)
        {
            int n = data.GetLength(0);
            int m = data.GetLength(1);
            double arg, cos, sin;
            Complex[] dst = new Complex[System.Math.Max(n, m)];

            // Verarbeite alle Zeilen.
            for (int i = 0; i < n; i++)
            {
                for (int j = 0; j < m; j++)
                {
                    dst[j] = Complex.Zero;

                    arg = -(int)direction * 2.0 * System.Math.PI * (double)j / (double)m;

                    // Summiere die Quellelemente.
                    for (int k = 0; k < m; k++)
                    {
                        cos = System.Math.Cos(k * arg);
                        sin = System.Math.Sin(k * arg);

                        dst[j].Re += (float)(data[i, k].Re * cos - data[i, k].Im * sin);
                        dst[j].Im += (float)(data[i, k].Re * sin + data[i, k].Im * cos);
                    }
                }

                // Kopiere die Elemente.
                if (direction == FourierDirection.Forward)
                {
                    for (int j = 0; j < m; j++)
                    {
                        data[i, j].Re = dst[j].Re / m;
                        data[i, j].Im = dst[j].Im / m;
                    }
                }
                else
                {
                    for (int j = 0; j < m; j++)
                    {
                        data[i, j].Re = dst[j].Re;
                        data[i, j].Im = dst[j].Im;
                    }
                }
            }

            // Verarbeite alle Spalten.
            for (int j = 0; j < m; j++)
            {
                for (int i = 0; i < n; i++)
                {
                    dst[i] = Complex.Zero;

                    arg = -(int)direction * 2.0 * System.Math.PI * (double)i / (double)n;

                    // Summiere die Quellelemente.
                    for (int k = 0; k < n; k++)
                    {
                        cos = System.Math.Cos(k * arg);
                        sin = System.Math.Sin(k * arg);

                        dst[i].Re += (float)(data[k, j].Re * cos - data[k, j].Im * sin);
                        dst[i].Im += (float)(data[k, j].Re * sin + data[k, j].Im * cos);
                    }
                }

                // Kopiere die Elemente.
                if (direction == FourierDirection.Forward)
                {
                    for (int i = 0; i < n; i++)
                    {
                        data[i, j].Re = dst[i].Re / n;
                        data[i, j].Im = dst[i].Im / n;
                    }
                }
                else
                {
                    for (int i = 0; i < n; i++)
                    {
                        data[i, j].Re = dst[i].Re;
                        data[i, j].Im = dst[i].Im;
                    }
                }
            }
        }

        /// <summary>
        /// Eindimensionale schnelle Fourier-Transformation.
        /// </summary>
        /// <param name="data">Zu transformierende Daten</param>
        /// <param name="direction">Transformationsrichtung</param>
        public static void FFT(Complex[] data, FourierDirection direction)
        {
            int n = data.Length;
            int m = Log2(n);

            // Umordnung der Daten.
            ReorderData(data);

            // FFT berechnen.
            int tn = 1, tm;

            for (int k = 1; k <= m; k++)
            {
                Complex[] rotation = GetComplexRotation(k, direction);

                tm = tn;
                tn <<= 1;

                for (int i = 0; i < tm; i++)
                {
                    Complex t = rotation[i];

                    for (int even = i; even < n; even += tn)
                    {
                        int odd = even + tm;
                        Complex ce = data[even];
                        Complex co = data[odd];

                        float tr = co.Re * t.Re - co.Im * t.Im;
                        float ti = co.Re * t.Im + co.Im * t.Re;

                        data[even].Re += tr;
                        data[even].Im += ti;

                        data[odd].Re = ce.Re - tr;
                        data[odd].Im = ce.Im - ti;
                    }
                }
            }

            if (direction == FourierDirection.Forward)
            {
                for (int i = 0; i < n; i++)
                {
                    data[i].Re /= (float)n;
                    data[i].Im /= (float)n;
                }
            }
        }

        /// <summary>
        /// Zweidimensionale schnelle Fourier-Transformation.
        /// </summary>
        /// <param name="data">Zu transformierende Daten</param>
        /// <param name="direction">Transformationsrichtung</param>
        public static void FFT2(Complex[,] data, FourierDirection direction)
        {
            int k = data.GetLength(0);
            int n = data.GetLength(1);

            // Gre der bergebenen Daten muss eine Zweierpotenz sein.
            if ((!IsPowerOf2(k)) || (!IsPowerOf2(n)) ||
                (k < minLength) || (k > maxLength) ||
                (n < minLength) || (n > maxLength))
            {
                throw new ArgumentException();
            }

            // Verarbeite alle Zeilen.
            Complex[] row = new Complex[n];
            for (int i = 0; i < k; i++)
            {
                // Kopiere die Zeile.
                for (int j = 0; j < n; j++)
                    row[j] = data[i, j];
                // Transformiere sie.
                FFT(row, direction);
                // Kopiere sie zurck.
                for (int j = 0; j < n; j++)
                    data[i, j] = row[j];
            }

            // Verarbeite alle Spalten.
            Complex[] col = new Complex[k];
            for (int j = 0; j < n; j++)
            {
                // Kopiere die Spalte.
                for (int i = 0; i < k; i++)
                    col[i] = data[i, j];
                // Transformiere sie.
                FFT(col, direction);
                // Kopiere sie zurck.
                for (int i = 0; i < k; i++)
                    data[i, j] = col[i];
            }
        }

        /// <summary>
        /// Berechnet die bergebene Zweierpotenz.
        /// </summary>
        /// <param name="exp">Exponent der zu berechnenden Potenz, exp \in [0..30]</param>
        /// <returns>Zweierpotenz des bergebenen Exponenten</returns>
        public static int Pow2(int exp)
        {
            return ((exp >= 0) && (exp <= 30)) ? (1 << exp) : 0;
        }

        /// <summary>
        /// Prft, ob die bergebene Zahl eine Zweierpotenz ist.
        /// </summary>
        /// <param name="x">Zu prfende Zahl</param>
        /// <returns>true, wenn x=2^n, false sonst</returns>
        public static bool IsPowerOf2(int x)
        {
            return (x & (x - 1)) == 0;
        }

        /// <summary>
        /// Gibt die Basis des dualen Logarithmus von x zurck.
        /// </summary>
        /// <param name="x">Argument, zu dem der ld berechnet werden soll</param>
        /// <returns>ld(x)</returns>
        public static int Log2(int x)
        {
            if (x <= 65536)
            {
                if (x <= 256)
                {
                    if (x <= 16)
                    {
                        if (x <= 4)
                        {
                            if (x <= 2)
                            {
                                if (x <= 1)
                                    return 0;
                                return 1;
                            }
                            return 2;
                        }
                        if (x <= 8)
                            return 3;
                        return 4;
                    }
                    if (x <= 64)
                    {
                        if (x <= 32)
                            return 5;
                        return 6;
                    }
                    if (x <= 128)
                        return 7;
                    return 8;
                }
                if (x <= 4096)
                {
                    if (x <= 1024)
                    {
                        if (x <= 512)
                            return 9;
                        return 10;
                    }
                    if (x <= 2048)
                        return 11;
                    return 12;
                }
                if (x <= 16384)
                {
                    if (x <= 8192)
                        return 13;
                    return 14;
                }
                if (x <= 32768)
                    return 15;
                return 16;
            }

            if (x <= 16777216)
            {
                if (x <= 1048576)
                {
                    if (x <= 262144)
                    {
                        if (x <= 131072)
                            return 17;
                        return 18;
                    }
                    if (x <= 524288)
                        return 19;
                    return 20;
                }
                if (x <= 4194304)
                {
                    if (x <= 2097152)
                        return 21;
                    return 22;
                }
                if (x <= 8388608)
                    return 23;
                return 24;
            }
            if (x <= 268435456)
            {
                if (x <= 67108864)
                {
                    if (x <= 33554432)
                        return 25;
                    return 26;
                }
                if (x <= 134217728)
                    return 27;
                return 28;
            }
            if (x <= 1073741824)
            {
                if (x <= 536870912)
                    return 29;
                return 30;
            }
            return 31;
        }
        #endregion

        #region Private Methoden

        private const int minLength = 2;
        private const int maxLength = 16384;
        private const int minBits = 1;
        private const int maxBits = 14;
        private static int[][] ReversedBits = new int[maxBits][];
        private static Complex[,][] complexRotation = new Complex[maxBits, 2][];

        // Erhalte ein Feld, das anzeigt welche Daten vor der FFT getauscht werden mssen.
        private static int[] GetReversedBits(int numberOfBits)
        {
            if ((numberOfBits < minBits) || (numberOfBits > maxBits))
                throw new ArgumentOutOfRangeException();

            // Ermittle, ob das Array bereits berechnet wurde.
            if (ReversedBits[numberOfBits - 1] == null)
            {
                int n = Pow2(numberOfBits);
                int[] rBits = new int[n];

                // Berechne das Array.
                for (int i = 0; i < n; i++)
                {
                    int oldBits = i;
                    int newBits = 0;

                    for (int j = 0; j < numberOfBits; j++)
                    {
                        newBits = (newBits << 1) | (oldBits & 1);
                        oldBits = (oldBits >> 1);
                    }
                    rBits[i] = newBits;
                }
                ReversedBits[numberOfBits - 1] = rBits;
            }
            return ReversedBits[numberOfBits - 1];
        }

        // Erhalte die Rotation der komplexen Zahlen
        private static Complex[] GetComplexRotation(int numberOfBits, FourierDirection diRection)
        {
            int directionIndex = (diRection == FourierDirection.Forward) ? 0 : 1;

            // Prfe, ob das Array bereits berechnet wurde.
            if (complexRotation[numberOfBits - 1, directionIndex] == null)
            {
                int n = 1 << (numberOfBits - 1);
                float uR = 1.0f;
                float uI = 0.0f;
                double angle = System.Math.PI / n * (int)diRection;
                float wR = (float)System.Math.Cos(angle);
                float wI = (float)System.Math.Sin(angle);
                float t;
                Complex[] rotation = new Complex[n];

                for (int i = 0; i < n; i++)
                {
                    rotation[i] = new Complex(uR, uI);
                    t = uR * wI + uI * wR;
                    uR = uR * wR - uI * wI;
                    uI = t;
                }

                complexRotation[numberOfBits - 1, directionIndex] = rotation;
            }
            return complexRotation[numberOfBits - 1, directionIndex];
        }

        // Ordne die Daten fr die FFT um.
        private static void ReorderData(Complex[] data)
        {
            int len = data.Length;

            // Prfe die Datenlnge.
            if ((len < minLength) || (len > maxLength) || (!IsPowerOf2(len)))
                throw new ArgumentException();

            int[] rBits = GetReversedBits(Log2(len));

            for (int i = 0; i < len; i++)
            {
                int s = rBits[i];

                if (s > i)
                {
                    Complex t = data[i];
                    data[i] = data[s];
                    data[s] = t;
                }
            }
        }

        // Fhre die Householder-Transformation durch, um die QR-Zerlegung zu berechnen.
        static private void house(double[,] A, double[] R)
        {
            int M = A.GetLength(0);
            int N = A.GetLength(1) - 1;
            double[] y = new double[N + 1];
            double kappa, s;
            for (int j = 0; j < N; j++)
            {
                // s berechnen.
                s = 0.0;
                for (int i = j; i < M; i++)
                    s += A[i, j] * A[i, j];
                if (s == 0.0) // A singulr?: wenn ja, dann wirf Ausnahme.
                {
                    throw new Exception("Singularittstest nicht bestanden.");
                }
                // Rho in R abspeichern.
                R[j] = (A[j, j] < 0) ? Math.Sqrt(s) : -Math.Sqrt(s);
                // Kappa berechnen.
                kappa = s - R[j] * A[j, j];
                A[j, j] -= R[j];
                for (int i = j + 1; i <= N; i++)
                {
                    y[i] = 0; // y berechnen.
                    for (int k = j; k < M; k++)
                        y[i] += A[k, j] * A[k, i];
                    y[i] /= kappa;
                    for (int k = j; k < M; k++) // Householder-transformierte Matrix A berechnen.
                        A[k, i] -= A[k, j] * y[i];
                }
            }
        }
        #endregion
    }
}
